#version 430

struct DirectionalLight {
    vec3 direction;
    vec3 intensity;
};

struct PointLight {
    vec3 position;
    vec3 intensity;
    vec3 attenuationFactors;
};

uniform float ambientStrength;

struct Material {
    sampler2D diffuse;
    sampler2D specular;
    sampler2D normal;
    bool hasNormalMap;
    float shininess;
};

in vec3 worldNormal;
in vec3 worldPosition;
in vec2 faceUv;
in vec4 FragPosLightSpace;
in mat3 TBN;
in vec3 TangentLightPos;
in vec3 TangentViewPos;
in vec3 TangentFragPos;

uniform samplerCube mysteryLightcubeCubemap;
uniform samplerCube fireflyLightobjectCubemap;

uniform vec3 mysteryLightcubePointLightPosition;
uniform vec3 fireflyLightobjectPointLightPosition;

uniform vec3 viewPosition;
uniform Material material;

uniform float far_plane;
uniform DirectionalLight directionalLights[6];
uniform uint directionalLightCount;

uniform PointLight pointLights[6];
uniform uint pointLightCount;


out vec4 fragColor;

float shadowCalculation(vec3 fragPos, vec3 worldNormalParam) {

    float shadow  = 0.0;
    float bias = max(0.05 * (1.0 - dot(worldNormalParam, fragPos)), 0.09);
    float samples = 4.0;
    float offset  = 0.1;
    vec3 fragToFireflyLight = fragPos - fireflyLightobjectPointLightPosition;
    vec3 fragToMysteryLight = fragPos - mysteryLightcubePointLightPosition;
    for(float x = -offset; x < offset; x += offset / (samples * 0.5))
    {
        for(float y = -offset; y < offset; y += offset / (samples * 0.5))
        {
            for(float z = -offset; z < offset; z += offset / (samples * 0.5))
            {
                float closestDepthMystery = texture(mysteryLightcubeCubemap, fragToMysteryLight + vec3(x, y, z)).r;
                closestDepthMystery *= far_plane;   // undo mapping [0;1]
                float currentDepthMystery = length(fragToMysteryLight);
                if(currentDepthMystery - bias > closestDepthMystery) {
                    shadow += 0.8;
                }
                float closestDepth = texture(fireflyLightobjectCubemap, fragToFireflyLight + vec3(x, y, z)).r;
                closestDepth *= far_plane;   // undo mapping [0;1]
                float currentDepth = length(fragToFireflyLight);
                if(currentDepth - bias > closestDepth) {
                    shadow += 0.2;
                }
            }
        }
    }
    return shadow /= (samples * samples * samples);
}



vec3 phongDirectionalIllumination(
    vec3 lightIntensity,
    vec3 lightDirection,
    float lightAttenuation,
    vec3 worldPosition,
    vec3 worldNormalParam
) {

    worldNormalParam = normalize(worldNormalParam);
    lightDirection = normalize(lightDirection);


    if (material.hasNormalMap) {
        // obtain normal from normal map in range [0,1]
        vec3 normal = texture(material.normal, faceUv).rgb;
        // transform normal vector to range [-1,1]
        normal = normalize(normal * 2.0 - 1.0);  // this normal is in tangent space
        worldNormalParam = normalize(TBN * normal);
    }

    vec3 ambientComponent = ambientStrength * texture(material.diffuse, faceUv).rgb;
    vec3 lightDir = lightDirection;
    float diff = max(dot(worldNormalParam, lightDir), 0.0);
    vec3 diffuseComponent = lightAttenuation * diff * texture(material.diffuse, faceUv).rgb * lightIntensity;

    vec3 viewDirection = normalize(viewPosition - worldPosition);
    vec3 reflectDirection = normalize(reflect(-lightDir, worldNormalParam));
    float spec = pow(max(dot(viewDirection, reflectDirection), 0.0), material.shininess);

    vec3 specularComponent = lightAttenuation * spec * texture(material.specular, faceUv).rgb * lightIntensity;
    specularComponent *= (1.0 - diff);
    float shadow = shadowCalculation(FragPosLightSpace.xyz, worldNormalParam);
    return ambientComponent + (1 - shadow) * (diffuseComponent  + specularComponent);
}

vec3 phongIllumination(vec3 worldPosition, vec3 worldNormalParam) {
    vec3 summedLight = vec3(0);

    for (int i = 0; i != min(directionalLightCount, 6); i++) {
        summedLight += phongDirectionalIllumination(
            directionalLights[i].intensity,
            -directionalLights[i].direction,
            1.0,
            worldPosition,
            worldNormalParam
        );
    }

    for (int i = 0; i != min(pointLightCount, 6); i++) {
        PointLight light = pointLights[i];
        vec3 direction = light.position - worldPosition;
        float dist = length(direction);
        float attenuation = 1.0 / (light.attenuationFactors.x + light.attenuationFactors.y * dist + light.attenuationFactors.z * dist * dist);

        summedLight += phongDirectionalIllumination(
            light.intensity,
            direction,
            attenuation,
            worldPosition,
            worldNormalParam
        );
    }

    return summedLight;
}

void main() {

    fragColor = vec4(phongIllumination(worldPosition, worldNormal), 1.0);



    // visualize depth map
    /*vec3 fragToLight = FragPosLightSpace.xyz - mysteryLightcubePointLightPosition;
    float closestDepth = texture(mysteryLightcubeCubemap, fragToLight).r;
    closestDepth *= far_plane;

   fragColor = vec4(vec3(closestDepth / far_plane), 1.0);*/
}
