#version 330 core

in vec2 TexCoords;

uniform sampler2D downsampledDepth;
uniform samplerCube shadowCubemap;
uniform float farPlane;
uniform vec3 cameraPosition;
uniform vec3 lightPosition;
uniform mat4 inverseViewProjection;
uniform float u_SceneTime; // Pass the scene time for dynamic noise

const float lightIntensity = 90.0;
const float density = 0.1;
const float decay = 0.99;
const float tau = 0.1;
const int maxSteps = 15;

float dither_pattern[16] = float[16](
    0.0f, 0.5f, 0.125f, 0.625f,
    0.75f, 0.22f, 0.875f, 0.375f,
    0.1875f, 0.6875f, 0.0625f, 0.5625,
    0.9375f, 0.4375f, 0.8125f, 0.3125
);

float dither_value = dither_pattern[(int(gl_FragCoord.x) % 4) * 4 + (int(gl_FragCoord.y) % 4)];

float tri(in float x) {
    return abs(fract(x) - 0.5);
}

vec3 tri3(in vec3 p) {
    return vec3(
        tri(p.z + tri(p.y * 1.)),
        tri(p.z + tri(p.x * 1.)),
        tri(p.y + tri(p.x * 1.))
    );
}

float triNoise3d(in vec3 p, in float spd, in float time) {
    float z = 1.4;
    float rz = 0.0;
    vec3 bp = p;
    for (float i = 0.; i <= 3.; i++) {
        vec3 dg = tri3(bp * 2.0);
        p += (dg + time * spd);
        bp *= 1.8;
        z *= 1.5;
        p *= 1.2;
        rz += (tri(p.z + tri(p.x + tri(p.y)))) / z;
        bp += 0.14;
    }
    return rz;
}

float visibilitySampling(vec3 rayPosition) {
    vec3 lightToFrag = rayPosition - lightPosition;
    float distance = length(lightToFrag);
    float storedDepth = texture(shadowCubemap, normalize(lightToFrag)).r;
    float currentDepth = distance / farPlane;
    return step(currentDepth, storedDepth + 0.0005);
}

out vec4 fragColor;

void main() {
    float fragDepth = texture(downsampledDepth, TexCoords).r;
    vec2 ndc = TexCoords * 2.0 - 1.0;
    float clipZ = fragDepth * 2.0 - 1.0;
    vec4 clipPos = vec4(ndc, clipZ, 1.0);
    vec4 fragWorldPosition = inverseViewProjection * clipPos;
    fragWorldPosition /= fragWorldPosition.w;

    vec3 rayDir = normalize(cameraPosition - fragWorldPosition.xyz);
    float stepSize = length(cameraPosition - fragWorldPosition.xyz) / maxSteps;
    vec3 rayPosition = fragWorldPosition.xyz;
    float accumulatedLight = 0.0;
    float distanceTraveled = 0.0;

    for (int i = 0; i < maxSteps; ++i) {
        rayPosition += rayDir * dither_value * stepSize;
        distanceTraveled += stepSize;

        float visibility = visibilitySampling(rayPosition);

        float distanceToLight = length(rayPosition - lightPosition);
        float attenuation = lightIntensity / (distanceToLight * distanceToLight);
        float scattering = exp(-tau * distanceToLight) * exp(-tau * distanceTraveled);

        float fog = triNoise3d(rayPosition * 2.2 / 8.0, 0.2, u_SceneTime);
        accumulatedLight += visibility * attenuation * scattering * stepSize + visibility * fog * 0.3;
    }

    accumulatedLight = accumulatedLight * (length(cameraPosition - fragWorldPosition.xyz) / farPlane) * 2;
    fragColor = vec4(accumulatedLight * density, accumulatedLight * density, accumulatedLight * density, 1.0);
}

