//
// Created by alina on 16.01.2025.
//

#ifndef AUDIO_H
#define AUDIO_H
#include <AL/al.h>
#include <AL/alc.h>
#include <sndfile/sndfile.h>
#include <vector>
#include <iostream>
#include <thread>

class Audio {
private:
    std::atomic<bool> stopFlag;
    std::atomic<glm::vec3> sourcePosition;
    std::atomic<glm::vec3> listenerPosition;
    std::thread workerThread;


    void playAudio(const char *filename) {

        SF_INFO sfInfo;
        SNDFILE *sndFile = sf_open(filename, SFM_READ, &sfInfo);
        if (!sndFile) {
            std::cerr << "Failed to open audio file: " << filename << std::endl;
            return;
        }
        std::vector<short> samples(sfInfo.frames * sfInfo.channels);
        sf_read_short(sndFile, samples.data(), samples.size());
        sf_close(sndFile);
        ALCdevice *device = alcOpenDevice(nullptr); // Open default device
        if (!device) {
            std::cerr << "Failed to open OpenAL device." << std::endl;
            return;
        }
        ALCcontext *context = alcCreateContext(device, nullptr);
        alcMakeContextCurrent(context);

        alDistanceModel(AL_INVERSE_DISTANCE_CLAMPED);

        ALuint buffer, source;
        alGenBuffers(1, &buffer);
        alGenSources(1, &source);

        // we assume a fixed source position, so we declare outside the loop
        ALfloat sourcePos[] = { sourcePosition.load().x, sourcePosition.load().y, sourcePosition.load().z };

        alSourcefv(source, AL_POSITION, sourcePos);
        alSourcef(source, AL_ROLLOFF_FACTOR, 5.0f);
        alSourcef(source, AL_REFERENCE_DISTANCE, 5.0f);
        alSourcef(source, AL_MAX_DISTANCE, 20.0f);

        alBufferData(buffer, AL_FORMAT_MONO16, samples.data(),
                     samples.size() * sizeof(short), sfInfo.samplerate);
        alSourcei(source, AL_BUFFER, buffer);
        alSourcePlay(source);
        std::cout << "Playing audio..." << std::endl;

        ALint state;
        do {
            ALfloat listenerPos[] = { listenerPosition.load().x, listenerPosition.load().y, listenerPosition.load().z };
            alListenerfv(AL_POSITION, listenerPos);
            alGetSourcei(source, AL_SOURCE_STATE, &state);
        } while (!stopFlag && state == AL_PLAYING);

        std::cout << "Stopping audio..." << std::endl;
        // Clean up
        alDeleteSources(1, &source);
        alDeleteBuffers(1, &buffer);
        alcMakeContextCurrent(nullptr);
        alcDestroyContext(context);
        alcCloseDevice(device);
    }

public:
    Audio(): stopFlag(false), workerThread(), listenerPosition(glm::vec3(0,0,0)), sourcePosition(glm::vec3(0,0,0)) {}

    void setListener(glm::vec3 pos) {
        listenerPosition.store(pos);
    }

    void setSource(glm::vec3 pos) {
        sourcePosition.store(pos);
    }

    void playAudioAsynchronous(const char *filename) {
        this->stopFlag.store(false);
         workerThread = std::thread([this, filename]() {
            this->playAudio(filename);
        });
    }

    void gracefullyStopAudio() {
        this->stopFlag.store(true);
        if (workerThread.joinable()) {
            workerThread.join(); // Wait for the thread to finish
        }
    }

    ~Audio() {
        gracefullyStopAudio();
    }
};


#endif //AUDIO_H
