#pragma once
#ifndef CAMERA_H
#define CAMERA_H

#include "PositionedModel.h"
#include <glm/ext.hpp>
#include <glm/gtx/string_cast.hpp>
#include <vector>
#include <glm/gtx/quaternion.hpp> 

class CameraKeyframe {

public:        
    glm::vec3 position;
    glm::vec3 forward;
    float transitionTime;

    CameraKeyframe(const glm::vec3& p, const glm::vec3& t, const float tt)
        : position(p), forward(t), transitionTime(tt) {}
};

class CatmullRomCamera {

public:
    glm::vec3 up = glm::vec3(0, 1, 0);
    glm::mat4 viewMatrix;
    glm::vec3 position;
    glm::vec3 forward;
    glm::mat4 projectionMatrix;
    std::vector<CameraKeyframe> keyframes;
    float yaw = -90.0f;
    float pitch = 0.0f;
    float movementSpeed = 3.0f;
    float mouseSensitivity = 0.1f;

    CatmullRomCamera(float fovy, float aspect, float zNear, float zFar)
        : projectionMatrix(glm::perspective(fovy, aspect, zNear, zFar)) {}

    void addKeyframe(const glm::vec3& position, const glm::vec3& target, const float transitionTime) {
        keyframes.emplace_back(CameraKeyframe(position, target, transitionTime));
    }
    struct TransformMatrices {
        glm::mat4 projectionViewMatrix;
        glm::mat4 modelMatrix;
    };

    TransformMatrices calcTransformationMatrices(const PositionedModel& model) const {
        glm::mat4 projectionViewMatrix = projectionMatrix * viewMatrix;
        return { projectionViewMatrix, model.modelMatrix() };
    }

    void CatmullRomCamera::processKeyboard(float deltaTime, bool* keys) {
        float velocity = movementSpeed * deltaTime;
        if (keys[GLFW_KEY_W])
            position += forward * velocity;
        if (keys[GLFW_KEY_S])
            position -= forward * velocity;
        if (keys[GLFW_KEY_A])
            position -= glm::normalize(glm::cross(forward, up)) * velocity;
        if (keys[GLFW_KEY_D])
            position += glm::normalize(glm::cross(forward, up)) * velocity;

        viewMatrix = glm::lookAt(position, position + forward, up);
    }

    void CatmullRomCamera::processMouseMovement(float xoffset, float yoffset) {
        xoffset *= mouseSensitivity;
        yoffset *= mouseSensitivity;

        yaw += xoffset;
        pitch += yoffset;

        if (pitch > 89.0f)
            pitch = 89.0f;
        if (pitch < -89.0f)
            pitch = -89.0f;

        glm::vec3 front;
        front.x = cos(glm::radians(pitch)) * cos(glm::radians(yaw));
        front.y = sin(glm::radians(pitch));
        front.z = cos(glm::radians(pitch)) * sin(glm::radians(yaw));
        forward = glm::normalize(front);

        viewMatrix = glm::lookAt(position, position + forward, up);
    }

    void interpolateOrJump(float currentTime) {
        if (keyframes.empty()) return;

        int k1, k2;
        float localTime;
        findKeyframes(currentTime, keyframes, k1, k2, localTime);

        if (keyframes[k2].transitionTime == 0) {
            glm::vec3 newPosition = keyframes[k2].position;
            glm::vec3 newForward = keyframes[k2].forward;
            updateCamera(newPosition, newForward);
        }
        else {
            interpolateCamera(currentTime, keyframes);
        }
    }

private:
    glm::vec3 catmullRomSpline(const glm::vec3& P0, const glm::vec3& P1, const glm::vec3& P2, const glm::vec3& P3, float t) {
        return 0.5f * ((2.0f * P1) + (-P0 + P2) * t + (2.0f * P0 - 5.0f * P1 + 4.0f * P2 - P3) * (t * t) + (-P0 + 3.0f * P1 - 3.0f * P2 + P3) * (t * t * t));
    }

    int findKeyframes(float currentTime, const std::vector<CameraKeyframe>& keyframes, int& k1, int& k2, float& localTime) {
        float accumulatedTime = 0.0f;

        for (int i = 0; i < keyframes.size() - 3; i++) {
            accumulatedTime += keyframes[i].transitionTime;

            if (currentTime < accumulatedTime) {
                k1 = i + 1;
                k2 = i + 2;
                localTime = (currentTime - (accumulatedTime - keyframes[i + 1].transitionTime)) / keyframes[i + 1].transitionTime;
                return 0;
            }
        }

        k1 = keyframes.size() - 3;
        k2 = keyframes.size() - 2;
        localTime = 1.0f;
        return 0;
    }

    void interpolateCamera(float currentTime, const std::vector<CameraKeyframe>& keyframes) {
        int k1, k2;
        float localTime;
        findKeyframes(currentTime, keyframes, k1, k2, localTime);

        CameraKeyframe kf0 = keyframes[k1 - 1];
        CameraKeyframe kf1 = keyframes[k1];
        CameraKeyframe kf2 = keyframes[k2];
        CameraKeyframe kf3 = keyframes[k2 + 1];

        glm::vec3 newPosition = catmullRomSpline(kf0.position, kf1.position, kf2.position, kf3.position, localTime);
        glm::vec3 newForward = catmullRomSpline(kf0.forward - newPosition, kf1.forward - newPosition, kf2.forward - newPosition, kf3.forward - newPosition, localTime);

        updateCamera(newPosition, newForward);
    }

    void updateCamera(const glm::vec3& newPosition, const glm::vec3& newForward) {
        position = newPosition;
        forward = glm::normalize(newForward);
        viewMatrix = glm::lookAt(newPosition, newForward + newPosition, up);
    }
};
#endif