#pragma once
#ifndef DRAWABLE_H
#define DRAWABLE_H

#include <assimp/Importer.hpp>
#include <assimp/scene.h>
#include <assimp/postprocess.h>
#include "TextureShadowCubemap.h"
#include "ShaderTexturePhong.h"
#include "ShaderShadowCubemap.h"
#include "Geometry.h"


class PointLight;

class Drawable : public PositionedModel {
public:
	Drawable(std::shared_ptr<ShaderProgram> sh, MaterialType mt)
		: shader{std::move(sh)} {
		materialType = mt;
	}

	void setMaterial(const PhongIlluminationMaterial &material) {
		illuminationMaterial = material;
	}

	void setDiffuseTexture(const std::shared_ptr<Texture> &shared) {
		diffuseTexture = shared;
	}

	void setSpecularTexture(const std::shared_ptr<Texture> &shared) {
		specularTexture = shared;
	}

	void scale(float x, float y, float z) {
		scaleVec = glm::vec4(x, y, z, 1);
	}

	virtual void draw(const CatmullRomCamera &camera, float far, const std::vector<DirectionalLight> &directionalLights,
	                  const std::vector<PointLight> &pointLights,
	                  std::shared_ptr<TextureShadowCubemap> mysteryLightcubeCubemapTexture,
	                  glm::vec3 mysteryLightcubePointLightPosition,
	                  std::shared_ptr<TextureShadowCubemap> fireflyLightobjectCubemapTexture,
	                  glm::vec3 fireflyLightobjectPointLightPosition
	) {
		auto transformMatrices = camera.calcTransformationMatrices(*this);
		for (auto geometry: this->geometries) {
			geometry->bind();
			std::shared_ptr<ShaderTexturePhong> textureShader = std::dynamic_pointer_cast<ShaderTexturePhong>(shader);
			textureShader->bind();
			textureShader->setMysteryLightcubeCubemap(*mysteryLightcubeCubemapTexture);
			textureShader->setFireflyLightobjectCubemap(*fireflyLightobjectCubemapTexture);

			textureShader->setMysteryLightcubePointLightPosition(mysteryLightcubePointLightPosition);
			textureShader->setFireflyLightobjectPointLightPosition(fireflyLightobjectPointLightPosition);

			textureShader->setFarPlane(far);
			textureShader->setViewPosition(camera.position);
			textureShader->setLightSpaceMatrix(lightSpaceMatrix);
			textureShader->setTransformationMatrices(transformMatrices);
			textureShader->setTexture(*diffuseTexture, *specularTexture, 0.8f);
			textureShader->setDirectionalLights(directionalLights);
			textureShader->setPointLights(pointLights);
			glDrawElements(GL_TRIANGLES, geometry->indexCount(), GL_UNSIGNED_INT, 0);
			ShaderProgram::unbind();
			Geometry::unbind();
		}
	}

	virtual void drawInLightPerspective(float far, const std::shared_ptr<ShaderShadowCubemap> &shader,
	                                    glm::vec3 pointLightPosition, std::vector<glm::mat4> cubefaceTransforms) {
		for (const auto &geometry: this->geometries) {
			geometry->bind();
			shader->setModelMatrix(modelMatrix());
			shader->setFarPlane(far);
			shader->setCubefaceTransforms(cubefaceTransforms);
			shader->setPointLightPosition(pointLightPosition);
			glDrawElements(GL_TRIANGLES, geometry->indexCount(), GL_UNSIGNED_INT, 0);
			Geometry::unbind();
		}
	}

protected:
	std::vector<std::shared_ptr<Geometry> > geometries;
	MaterialType materialType;
	std::shared_ptr<ShaderProgram> shader;
	PhongIlluminationMaterial illuminationMaterial;
	std::shared_ptr<TextureShadowCubemap> shadowMapTexture;
	std::shared_ptr<Texture> diffuseTexture;
	std::shared_ptr<Texture> specularTexture;
	glm::mat4 lightSpaceMatrix = glm::mat4(1.0);

private:
	glm::mat4 rotationMatrix = glm::mat4(1.0);
	glm::mat4 translationMatrix = glm::mat4(1.0);
	glm::vec4 scaleVec = glm::vec4(1.0f);
};

class LoadedModel final : public Drawable {
public:
	LoadedModel(const std::shared_ptr<ShaderTexturePhong> &sh, const std::string &modelName,
	            const std::string &modelDirectory)
		: Drawable(sh, MaterialType::TEXTURE) {
		textureShader = sh;
		diffuseTexture = Texture::create("assets/textures/default.jpg");
		specularTexture = Texture::create("assets/textures/default_dark.jpg");
		loadModel(modelDirectory + "/" + modelName, modelDirectory);
	}

private:
	std::shared_ptr<ShaderTexturePhong> textureShader;
	std::vector<std::shared_ptr<Texture> > textures_loaded;

	void loadModel(const std::string &modelPath, const std::string &modelDirectory) {
		Assimp::Importer importer;

		const aiScene *scene = importer.ReadFile(
			modelPath,
			aiProcess_Triangulate | aiProcess_GenSmoothNormals | aiProcess_FlipUVs | aiProcess_CalcTangentSpace);

		if (!scene || scene->mFlags & AI_SCENE_FLAGS_INCOMPLETE || !scene->mRootNode) {
			throw std::runtime_error("Failed to load model: " + modelPath);
		}

		for (unsigned int j = 0; j < scene->mNumMeshes; j++) {
			const aiMesh *mesh = scene->mMeshes[j];

			std::vector<glm::vec3> vertices;
			std::vector<GLuint> indices;
			std::vector<glm::vec2> uvs;

			for (unsigned int i = 0; i < mesh->mNumVertices; i++) {
				aiVector3D aiVertex = mesh->mVertices[i];
				aiVector3D aiNormal = mesh->mNormals[i];
				aiVector3D aiTangents = mesh->mTangents[i];
				aiVector3D aiUV = mesh->mTextureCoords[0] ? mesh->mTextureCoords[0][i] : aiVector3D(0.0f, 0.0f, 0.0f);


				vertices.push_back(glm::vec3(aiVertex.x, aiVertex.y, aiVertex.z));
				vertices.push_back(glm::vec3(aiNormal.x, aiNormal.y, aiNormal.z));
				vertices.push_back(glm::vec3(aiTangents.x, aiTangents.y, aiTangents.z));
				uvs.push_back(glm::vec2(aiUV.x, aiUV.y));
			}

			for (unsigned int i = 0; i < mesh->mNumFaces; i++) {
				const aiFace &face = mesh->mFaces[i];
				for (unsigned int k = 0; k < face.mNumIndices; k++) {
					indices.push_back(face.mIndices[k]);
				}
			}

			auto geometry = std::make_shared<Geometry>();
			geometry->init(shader->attributeLocation("position", 0),
			               shader->attributeLocation("normal", 1),
			               shader->attributeLocation("uv", 2),
			               shader->attributeLocation("tangent", 3));
			geometry->setData(vertices, indices, uvs);
			geometries.push_back(geometry);

			aiMaterial *material = scene->mMaterials[mesh->mMaterialIndex];
			auto diffuseMaps = loadMaterialTextures(material, aiTextureType_DIFFUSE, "texture_diffuse", modelDirectory);
			geometry->addDiffuseTextures(diffuseMaps);
			auto specularMaps = loadMaterialTextures(material, aiTextureType_SPECULAR, "texture_specular",
			                                         modelDirectory);
			geometry->addSpecularTextures(specularMaps);
			auto normalMaps = loadMaterialTextures(material, aiTextureType_NORMALS, "texture_normal", modelDirectory);
			geometry->addNormalTextures(normalMaps);
		}
	}

	std::vector<std::shared_ptr<Texture> > loadMaterialTextures(aiMaterial *mat, aiTextureType type,
	                                                            const std::string &typeName,
	                                                            const std::string &modelDirectory) {
		std::vector<std::shared_ptr<Texture> > textures;
		for (unsigned int i = 0; i < mat->GetTextureCount(type); i++) {
			aiString str;
			mat->GetTexture(type, i, &str);
			bool skip = false;
			for (auto &loadedTexture: textures_loaded) {
				if (std::strcmp(loadedTexture->path.data(), str.C_Str()) == 0) {
					textures.push_back(loadedTexture);
					skip = true;
					break;
				}
			}
			if (!skip) {
				std::string filename = modelDirectory + "/../" + str.C_Str();
				auto texture = Texture::create(filename.c_str());
				texture->type = typeName;
				textures.push_back(std::move(texture));
				textures_loaded.push_back(textures.back());
			}
		}
		return textures;
	}

	void draw(const CatmullRomCamera &camera, const float far, const std::vector<DirectionalLight> &directionalLights,
	          const std::vector<PointLight> &pointLights,
	          std::shared_ptr<TextureShadowCubemap> mysteryLightcubeCubemapTexture,
	          glm::vec3 mysteryLightcubePointLightPosition,
	          std::shared_ptr<TextureShadowCubemap> fireflyLightobjectCubemapTexture,
	          glm::vec3 fireflyLightobjectPointLightPosition
	) override {
		auto transformMatrices = camera.calcTransformationMatrices(*this);

		textureShader->bind();
		textureShader->setMysteryLightcubeCubemap(*mysteryLightcubeCubemapTexture);
		textureShader->setFireflyLightobjectCubemap(*fireflyLightobjectCubemapTexture);

		textureShader->setMysteryLightcubePointLightPosition(mysteryLightcubePointLightPosition);
		textureShader->setFireflyLightobjectPointLightPosition(fireflyLightobjectPointLightPosition);

		textureShader->setFarPlane(far);

		textureShader->setAmbientStrength(0.1f);
		textureShader->setViewPosition(camera.position);
		textureShader->setLightSpaceMatrix(lightSpaceMatrix);
		textureShader->setTransformationMatrices(transformMatrices);
		textureShader->setDirectionalLights(directionalLights);
		textureShader->setPointLights(pointLights);

		for (unsigned int i = 0; i < geometries.size(); i++) {
			auto geometry = geometries[i];
			geometry->bind();

			if (geometry->diffuseTextures.size() > 0) {
				if (geometry->specularTextures.size() > 0) {
					textureShader->addTexture(*geometry->diffuseTextures[0], *geometry->specularTextures[0], 0.2f, i);
				} else {
					textureShader->addTexture(*geometry->diffuseTextures[0], *specularTexture, 0.2f, i);
				}
			} else {
				textureShader->setTexture(*diffuseTexture, *specularTexture, 0.2f);
			}

			if (geometry->normalTextures.size() > 0) {
				textureShader->setNormalMap(*geometry->normalTextures[0]);
				textureShader->setHasNormalMap(true);
			} else {
				textureShader->setHasNormalMap(false);
			}

			glDrawElements(GL_TRIANGLES, geometries[i]->indexCount(), GL_UNSIGNED_INT, 0);

			Geometry::unbind();
		}

		ShaderProgram::unbind();
	}
};
#endif
