#pragma once
#ifndef GEOMETRY_H
#define GEOMETRY_H

#include "Bindable.h"
#include "Buffer.h"
#include "Utils.h"
#include "Shader.h"
#include <vector>

class Geometry : public Bindable {
public:
	std::vector<std::shared_ptr<Texture> > diffuseTextures;
	std::vector<std::shared_ptr<Texture> > specularTextures;
	std::vector<std::shared_ptr<Texture> > normalTextures;

	Geometry() {
		glGenVertexArrays(1, &vaoHandle);
	}

	virtual ~Geometry() {
		glDeleteVertexArrays(1, &vaoHandle);
	}

	virtual void bind() const override {
		glBindVertexArray(vaoHandle);
	}

	void init(const ShaderAttribute &position, const ShaderAttribute &normal, const ShaderAttribute &uv) {
		if (!position) {
			throw std::runtime_error("Invalid 'postion' attribute for geometry object init");
		}

		if (!normal) {
			throw std::runtime_error("Invalid 'normal' attribute for geometry object init");
		}

		if (!uv) {
			throw std::runtime_error("Invalid 'uv' attribute for geometry object init");
		}

		hasTangents = false;
		bind();

		vertecies.bind();
		constexpr auto vertexSize = 2 * sizeof(glm::vec3);
		std::size_t offset = 0;
		glEnableVertexAttribArray(position.handle());
		glVertexAttribPointer(position.handle(), 3, GL_FLOAT, GL_FALSE, vertexSize, (void *) offset);
		offset += 1 * sizeof(glm::vec3);
		glEnableVertexAttribArray(normal.handle());
		glVertexAttribPointer(normal.handle(), 3, GL_FLOAT, GL_FALSE, vertexSize, (void *) offset);
		offset += 1 * sizeof(glm::vec3);

		uvs.bind();
		glEnableVertexAttribArray(uv.handle());
		glVertexAttribPointer(uv.handle(), 2, GL_FLOAT, GL_FALSE, 1 * sizeof(glm::vec2), (void *) 0);

		indices.bind();

		unbind();
	}

	void init(const ShaderAttribute &position, const ShaderAttribute &normal, const ShaderAttribute &uv,
	          const ShaderAttribute &tangent) {
		if (!position) {
			throw std::runtime_error("Invalid 'postion' attribute for geometry object init");
		}

		if (!normal) {
			throw std::runtime_error("Invalid 'normal' attribute for geometry object init");
		}

		if (!uv) {
			throw std::runtime_error("Invalid 'uv' attribute for geometry object init");
		}

		if (!tangent) {
			throw std::runtime_error("Invalid 'tangent' attribute for geometry object init");
		}
		hasTangents = true;
		bind();

		vertecies.bind();
		constexpr auto vertexSize = 3 * sizeof(glm::vec3);
		std::size_t offset = 0;
		glEnableVertexAttribArray(position.handle());
		glVertexAttribPointer(position.handle(), 3, GL_FLOAT, GL_FALSE, vertexSize, (void *) offset);
		offset += 1 * sizeof(glm::vec3);
		glEnableVertexAttribArray(normal.handle());
		glVertexAttribPointer(normal.handle(), 3, GL_FLOAT, GL_FALSE, vertexSize, (void *) offset);
		offset += 1 * sizeof(glm::vec3);

		glEnableVertexAttribArray(tangent.handle());
		glVertexAttribPointer(tangent.handle(), 3, GL_FLOAT, GL_FALSE, vertexSize, (void *) offset);


		uvs.bind();
		glEnableVertexAttribArray(uv.handle());
		glVertexAttribPointer(uv.handle(), 2, GL_FLOAT, GL_FALSE, 1 * sizeof(glm::vec2), (void *) 0);

		indices.bind();

		unbind();
	}

	void setData(const std::vector<glm::vec3> &vertexData, const std::vector<GLuint> indexData,
	             const std::vector<glm::vec2> &uvData) {
		if (!hasTangents) {
			if (vertexData.size() % 2 != 0) {
				throw std::runtime_error("Number of vertecies not divisble by two");
			}

			if ((vertexData.size() / 2) != uvData.size()) {
				throw std::runtime_error("Number of vertecies not equals to the number of uvs");
			}

			vertecies.setData(vertexData.data(), vertexData.size() * sizeof(glm::vec3));
			uvs.setData(uvData.data(), uvData.size() * sizeof(glm::vec2));
			indices.setData(indexData.data(), indexData.size() * sizeof(GLuint));
			storedIndices = indexData.size();
		} else {
			if (vertexData.size() % 3 != 0) {
				throw std::runtime_error("Number of vertecies not divisble by 3");
			}

			if ((vertexData.size() / 3) != uvData.size()) {
				throw std::runtime_error("Number of vertecies not equals to the number of uvs");
			}

			vertecies.setData(vertexData.data(), vertexData.size() * sizeof(glm::vec3));
			uvs.setData(uvData.data(), uvData.size() * sizeof(glm::vec2));
			indices.setData(indexData.data(), indexData.size() * sizeof(GLuint));
			storedIndices = indexData.size();
		}
	}

	void setData(const glm::vec3 *vertexData, unsigned int vertexCount, const GLuint *indexData,
	             unsigned int indexCount, const glm::vec2 *uvData, unsigned int uvCount) {
		if (vertexCount % 2 != 0) {
			throw std::runtime_error("Number of vertecies not divisble by two");
		}

		if ((vertexCount / 2) != uvCount) {
			throw std::runtime_error("Number of vertecies not equals to the number of uvs");
		}

		vertecies.setData(vertexData, vertexCount * sizeof(glm::vec3));
		uvs.setData(uvData, uvCount * sizeof(glm::vec2));
		indices.setData(indexData, indexCount * sizeof(GLuint));
		storedIndices = indexCount;
	}

	auto indexCount() const {
		return storedIndices;
	}


	void addDiffuseTextures(const std::vector<std::shared_ptr<Texture> > &textures) {
		diffuseTextures.insert(diffuseTextures.end(), textures.begin(), textures.end());
	}

	void addSpecularTextures(const std::vector<std::shared_ptr<Texture> > &textures) {
		specularTextures.insert(specularTextures.end(), textures.begin(), textures.end());
	}

	void addNormalTextures(const std::vector<std::shared_ptr<Texture> > &textures) {
		normalTextures.insert(normalTextures.end(), textures.begin(), textures.end());
	}

	static void unbind() {
		glBindVertexArray(0);
		VertexBuffer::unbind();
		IndexBuffer::unbind();
	}

private:
	GLuint vaoHandle;
	VertexBuffer vertecies;
	VertexBuffer uvs;
	IndexBuffer indices;
	unsigned int storedIndices;
	bool hasTangents;
};

#endif
