// Local includes
#include "Utils.h"
#include "Window.h"
#include "Camera.h"
#include "Texture.h"
#include "Geometry.h"
#include "Materials.h"
#include "PointLight.h"
#include "Drawable.h"
#include "Scene.h"
#include "TextureSceneColor.h"
#include "TextureSceneColor.h"
#include "ShaderVolumetricGodRay.h"
#include "ShaderShadowCubemap.h"

// External includes
#include <glm/ext.hpp>
#include <glm/gtx/string_cast.hpp>

// Standard library includes
#include <optional>
#include <vector>
#include <random>

#include "Audio.h"
#include "ShaderDepthDownsampling.h"
#include "ShaderLightUpsampling.h"
#include "ShaderBilateralBlur.h"

// Template function to format debug message 
template<typename T>
T &glPrintFormattedDebugMessage(T &stream, GLenum source, GLenum type, GLuint id, GLenum severity,
                                const GLchar *message) {
    stream << "OpenGL Error: " << message << " [Source = ";

    switch (source) {
        case GL_DEBUG_SOURCE_API: stream << "API";
            break;
        case GL_DEBUG_SOURCE_APPLICATION: stream << "Application";
            break;
        case GL_DEBUG_SOURCE_WINDOW_SYSTEM: stream << "Window System";
            break;
        case GL_DEBUG_SOURCE_SHADER_COMPILER: stream << "Shader Compiler";
            break;
        case GL_DEBUG_SOURCE_THIRD_PARTY: stream << "Third Party";
            break;
        case GL_DEBUG_SOURCE_OTHER: stream << "Other";
            break;
        default: stream << "Unknown";
            break;
    }

    stream << ", Type = ";
    switch (type) {
        case GL_DEBUG_TYPE_ERROR: stream << "Error";
            break;
        case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR: stream << "Deprecated Behavior";
            break;
        case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR: stream << "Undefined Behavior";
            break;
        case GL_DEBUG_TYPE_PORTABILITY_ARB: stream << "Portability";
            break;
        case GL_DEBUG_TYPE_PERFORMANCE: stream << "Performance";
            break;
        case GL_DEBUG_TYPE_OTHER: stream << "Other";
            break;
        default: stream << "Unknown";
            break;
    }

    stream << ", Severity = ";
    switch (severity) {
        case GL_DEBUG_SEVERITY_HIGH: stream << "High";
            break;
        case GL_DEBUG_SEVERITY_MEDIUM: stream << "Medium";
            break;
        case GL_DEBUG_SEVERITY_LOW: stream << "Low";
            break;
        default: stream << "Unknown";
            break;
    }

    stream << ", ID = " << id << "]";
    return stream;
}

// Opengl debug callback forwarding message to template function
void APIENTRY glDebugCallback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length,
                              const GLchar *message, const GLvoid *userParam) {
    // Ignore nvidia performance warnings
    if (id == 131185 || id == 131218) {
        return;
    }
    // Ignore renderdoc debug group warnings
    if (type == GL_DEBUG_TYPE_PUSH_GROUP ||
        type == GL_DEBUG_TYPE_POP_GROUP ||
        type == GL_DEBUG_TYPE_MARKER) {
        return;
    }
    glPrintFormattedDebugMessage(std::cerr, source, type, id, severity, message) << "\n";
}

int main(int argc, char **argv) {
    // Read project settings from .ini
    INIReader reader("assets/settings.ini");
    int windowWidth = reader.GetInteger("window", "width", 800);
    int windowHeight = reader.GetInteger("window", "height", 800);
    std::string windowTitle = reader.Get("window", "title", "EZG23");
    float cameraFovy = glm::radians(reader.GetReal("camera", "fovy", 60.0));
    float depthCubemapTextureWidth = reader.GetReal("depth_cubemap_texture", "width", 1024);
    float depthCubemapTextureHeight = reader.GetReal("depth_cubemap_texture", "height", 1024);
    bool wireframeRendering = reader.GetBoolean("render_config", "wireframe_rendering", false);
    bool backfaceCulling = reader.GetBoolean("render_config", "backface_culling", false);
    bool glewExperimental = reader.GetBoolean("render_config", "glew_experimantal", true);
    float nearPlane = reader.GetReal("camera", "near", 0.01);
    float farPlane = reader.GetReal("camera", "far", 100.0);
    int downsamplingFactor = 3;

    // Glfw
    auto glfwInitStatus = glfwInit();
    if (glfwInitStatus == false) {
        EXIT_WITH_ERROR("GLFW initialization error");
    }

    // Glfw: Register error callback
    glfwSetErrorCallback([](int error, const char *description) {
        std::cerr << "GLFW error: " << description << std::endl;
    });

    // Glfw: Create window and bind as current context
    auto window = Window::create(windowWidth, windowHeight, std::move(windowTitle));
    if (!window) EXIT_WITH_ERROR("Window::create error");
    window->bind();

    // Glew
    auto glewInitStatus = glewInit();
    if (glewInitStatus != GLEW_OK) {
        std::cerr << "GLEW initialization errpr: " << glewGetErrorString(glewInitStatus) << std::endl;
        EXIT_WITH_ERROR("GLEW error");
    }

    // Framework
    auto frameworkInitStatus = initFramework();
    if (frameworkInitStatus == false) {
        EXIT_WITH_ERROR("Framework init error");
    }

    // Register opengl debug callback
#if _DEBUG
    glDebugMessageCallback(glDebugCallback, NULL);
    glEnable(GL_DEBUG_OUTPUT_SYNCHRONOUS);
#endif

    // Enable depth testing (culling by z-value in rasterization)
    glEnable(GL_DEPTH_TEST);

    // Shaders
    std::shared_ptr<ShaderTexturePhong> texturePhongShader = ShaderProgram::createFromFileName<ShaderTexturePhong>(
        "assets/shaders/texturePhong", false);
    if (!texturePhongShader) {
        EXIT_WITH_ERROR("ShaderProgram::createFromFileName<TexturePhongShader> error");
    }

    std::shared_ptr<ShaderShadowCubemap> shadowCubemapShader = ShaderProgram::createFromFileName<ShaderShadowCubemap>(
        "assets/shaders/shadowCubemap", true);
    if (!shadowCubemapShader) {
        EXIT_WITH_ERROR("ShaderProgram::createFromFileName<ShadowCubeMapShader> error");
    }

    std::shared_ptr<ShaderVolumetricGodray> volumetricGodRayShader = ShaderProgram::createFromFileName<
        ShaderVolumetricGodray>("assets/shaders/volumetricGodRay", false);
    if (!volumetricGodRayShader) {
        EXIT_WITH_ERROR("ShaderProgram::createFromFileName<VolumetricGodRayShader> error");
    }

    std::shared_ptr<ShaderDepthDownsampling> depthDownsampleShader = ShaderProgram::createFromFileName<
        ShaderDepthDownsampling>("assets/shaders/depthDownsampling", false);
    if (!depthDownsampleShader) {
        EXIT_WITH_ERROR("Failed to load DepthDownsampleShader");
    }

    std::shared_ptr<ShaderLightUpsampling> lightUpsampleShader = ShaderProgram::createFromFileName<
        ShaderLightUpsampling>("assets/shaders/lightUpsampling", false);
    if (!lightUpsampleShader) {
        EXIT_WITH_ERROR("Failed to load LightUpsampleShader");
    }

    std::shared_ptr<ShaderBilateralBlur> bilateralBlurShader = ShaderProgram::createFromFileName<ShaderBilateralBlur>(
        "assets/shaders/bilateralBlur", false);
    if (!bilateralBlurShader) {
        EXIT_WITH_ERROR("Failed to load BilateralBlurShader");
    }

    // Textures
    std::shared_ptr<TextureShadowCubemap> mysteryLightcubeCubemapTexture = TextureShadowCubemap::create(
        depthCubemapTextureWidth, depthCubemapTextureHeight);
    if (!mysteryLightcubeCubemapTexture) {
        EXIT_WITH_ERROR("ShadowCubeMapTexture::create error");
    }

    std::shared_ptr<TextureShadowCubemap> fireflyLightobjectCubemapTexture = TextureShadowCubemap::create(
        depthCubemapTextureWidth, depthCubemapTextureHeight);
    if (!fireflyLightobjectCubemapTexture) {
        EXIT_WITH_ERROR("ShadowCubeMapTexture::create error");
    }

    std::shared_ptr<TextureSceneDepth> sceneDepthTexture = TextureSceneDepth::create(windowWidth, windowHeight);
    if (!sceneDepthTexture) {
        EXIT_WITH_ERROR("SceneDepthTexture::create error");
    }

    std::shared_ptr<TextureSceneColor> sceneColorTexture = TextureSceneColor::create(windowWidth, windowHeight);
    if (!sceneColorTexture) {
        EXIT_WITH_ERROR("SceneColorTexture::create error");
    }

    std::shared_ptr<TextureSceneDepth> downsampledDepthTexture = TextureSceneDepth::create(
        windowWidth / downsamplingFactor, windowHeight / downsamplingFactor);
    if (!downsampledDepthTexture) {
        EXIT_WITH_ERROR("Failed to create downsampled depth texture");
    }

    std::shared_ptr<TextureSceneColor> upsampledLightTexture = TextureSceneColor::create(windowWidth, windowHeight);
    if (!upsampledLightTexture) {
        EXIT_WITH_ERROR("Failed to create upsampled light texture");
    }

    std::shared_ptr<TextureSceneColor> downsampledGodRayTexture = TextureSceneColor::create(
        windowWidth / downsamplingFactor, windowHeight / downsamplingFactor);
    if (!downsampledGodRayTexture) {
        EXIT_WITH_ERROR("Failed to create upsampled light texture");
    }

    std::shared_ptr<TextureSceneColor> blurredGodRayTexture = TextureSceneColor::create(
        windowWidth / downsamplingFactor, windowHeight / downsamplingFactor);
    if (!blurredGodRayTexture) {
        EXIT_WITH_ERROR("Failed to create blurred god ray texture");
    }

    // FBOs
    unsigned int mysteryLightcubeCubemapFBO;
    glGenFramebuffers(1, &mysteryLightcubeCubemapFBO);
    mysteryLightcubeCubemapTexture->attachToFrameBuffer(mysteryLightcubeCubemapFBO);

    unsigned int fireflyLightobjectCubemapFBO;
    glGenFramebuffers(1, &fireflyLightobjectCubemapFBO);
    fireflyLightobjectCubemapTexture->attachToFrameBuffer(fireflyLightobjectCubemapFBO);

    unsigned int sceneColorDepthFBO;
    glGenFramebuffers(1, &sceneColorDepthFBO);
    sceneDepthTexture->attachToFrameBuffer(sceneColorDepthFBO);
    sceneColorTexture->attachToFrameBuffer(sceneColorDepthFBO);

    unsigned int downsampleDepthFBO;
    glGenFramebuffers(1, &downsampleDepthFBO);
    downsampledDepthTexture->attachToFrameBuffer(downsampleDepthFBO);

    unsigned int volumetricGodRayFBO;
    glGenFramebuffers(1, &volumetricGodRayFBO);
    downsampledGodRayTexture->attachToFrameBuffer(volumetricGodRayFBO);

    unsigned int upsampleLightFBO;
    glGenFramebuffers(1, &upsampleLightFBO);
    upsampledLightTexture->attachToFrameBuffer(upsampleLightFBO);

    unsigned int bilateralBlurFBO;
    glGenFramebuffers(1, &bilateralBlurFBO);
    blurredGodRayTexture->attachToFrameBuffer(bilateralBlurFBO);

    // CatmullRomCamera
    CatmullRomCamera camera = CatmullRomCamera{cameraFovy, (1.0f * windowWidth) / windowHeight, nearPlane, farPlane};

    // CatmulRomCamera: Keyframe animate flying camera
    camera.addKeyframe(glm::vec3(-0.152449, 2.098354, 6.046732), glm::vec3(-0.138498, 2.129764, 5.047323), 2.0f);
    camera.addKeyframe(glm::vec3(-0.152449, 2.098354, 6.046732), glm::vec3(-0.138498, 2.129764, 5.047323), 2.0f);
    camera.addKeyframe(glm::vec3(-0.152449, 2.098354, 6.046732), glm::vec3(-0.138498, 2.129764, 5.047323), 2.0f);
    camera.addKeyframe(glm::vec3(-0.074876, 2.267675, -0.104539), glm::vec3(-0.057430, 2.260694, -1.104362), 2.0f);
    camera.addKeyframe(glm::vec3(1.278962, 2.322945, -6.835819), glm::vec3(1.176236, 2.284557, -7.829788), 2.0f);
    camera.addKeyframe(glm::vec3(0.955499, 1.866618, -13.869454), glm::vec3(0.842538, 1.616238, -14.830990), 2.0f);
    camera.addKeyframe(glm::vec3(0.254770, -0.262507, -20.607721), glm::vec3(0.056938, -0.482353, -21.562986), 2.0f);
    camera.addKeyframe(glm::vec3(-0.979674, -1.917704, -27.106735), glm::vec3(-1.671479, -2.067239, -27.813168), 2.0f);
    camera.addKeyframe(glm::vec3(-5.429423, -2.449518, -30.983597), glm::vec3(-6.412954, -2.534935, -31.142881), 2.0f);
    camera.addKeyframe(glm::vec3(-12.407818, -2.946474, -31.840349), glm::vec3(-13.395751, -3.099460, -31.816208),
                       2.0f);
    camera.addKeyframe(glm::vec3(-20.920179, -3.013415, -32.812714), glm::vec3(-21.900745, -3.199082, -32.749313),
                       2.0f);
    camera.addKeyframe(glm::vec3(-28.572241, -4.661573, -32.846729), glm::vec3(-29.553596, -4.835222, -32.764336),
                       2.0f);
    camera.addKeyframe(glm::vec3(-36.432610, -6.043537, -31.724590), glm::vec3(-37.416939, -6.205141, -31.654043),
                       2.0f);
    camera.addKeyframe(glm::vec3(-43.272583, -7.113265, -32.879921), glm::vec3(-44.241447, -7.250709, -32.673985),
                       2.0f);
    camera.addKeyframe(glm::vec3(-49.057709, -7.635369, -36.828461), glm::vec3(-49.789593, -7.779725, -36.162498),
                       2.0f);
    camera.addKeyframe(glm::vec3(-55.152813, -7.950152, -40.466969), glm::vec3(-55.343819, -8.075485, -39.493412),
                       2.0f);
    camera.addKeyframe(glm::vec3(-62.310902, -6.119682, -40.857613), glm::vec3(-61.925377, -6.375127, -39.970982),
                       2.0f);
    camera.addKeyframe(glm::vec3(-68.773422, -3.422158, -41.560181), glm::vec3(-68.132172, -3.752672, -40.867676),
                       2.0f);
    camera.addKeyframe(glm::vec3(-72.821648, -1.446352, -35.691093), glm::vec3(-71.987869, -1.864219, -35.330257),
                       2.0f);
    camera.addKeyframe(glm::vec3(-71.240166, -1.821986, -28.697517), glm::vec3(-70.352837, -2.282186, -28.726929),
                       2.0f);
    camera.addKeyframe(glm::vec3(-65.993820, -3.980456, -23.019352), glm::vec3(-65.247787, -4.403074, -23.533962),
                       2.0f);
    camera.addKeyframe(glm::vec3(-59.257153, -7.191608, -24.544651), glm::vec3(-58.822247, -7.518826, -25.383570),
                       2.0f);
    camera.addKeyframe(glm::vec3(-52.929413, -8.782808, -26.769979), glm::vec3(-53.745441, -8.854305, -27.343554),
                       2.0f);
    camera.addKeyframe(glm::vec3(-51.438679, -8.285531, -33.716606), glm::vec3(-52.242363, -8.400468, -33.132755),
                       2.0f);
    camera.addKeyframe(glm::vec3(-55.536110, -5.970026, -40.310966), glm::vec3(-55.693668, -6.269065, -39.369823),
                       2.0f);
    camera.addKeyframe(glm::vec3(-61.830181, -2.518746, -42.491241), glm::vec3(-61.534451, -2.964943, -41.646580),
                       2.0f);
    camera.addKeyframe(glm::vec3(-67.451187, -0.083422, -43.435947), glm::vec3(-66.947792, -0.554433, -42.711555),
                       2.0f);
    camera.addKeyframe(glm::vec3(-67.451187, -0.083422, -43.435947), glm::vec3(-66.947792, -0.554433, -42.711555),
                       2.0f);
    camera.addKeyframe(glm::vec3(-67.451187, -0.083422, -43.435947), glm::vec3(-66.947792, -0.554433, -42.711555),
                       2.0f);

    // Models
    auto caveModel = std::make_unique<LoadedModel>(std::move(texturePhongShader), "cave.obj", "assets/models");
    if (!caveModel) {
        EXIT_WITH_ERROR("std::make_unique<LoadedModel> error: cave.obj");
    }
    auto mysteryLightcubeModel = std::make_unique<LoadedModel>(std::move(texturePhongShader), "mystery_cube.obj",
                                                               "assets/models");
    if (!caveModel) {
        EXIT_WITH_ERROR("std::make_unique<LoadedModel> error: mysterycube.obj");
    }
    auto fireflyLightobjectModel = std::make_unique<LoadedModel>(std::move(texturePhongShader), "star.obj",
                                                                 "assets/models");
    if (!fireflyLightobjectModel) {
        EXIT_WITH_ERROR("std::make_unique<DrawableSphere> error");
    }
    auto polyhedra1Model = std::make_unique<LoadedModel>(std::move(texturePhongShader), "polyhedra_1.obj",
                                                         "assets/models");
    if (!polyhedra1Model) {
        EXIT_WITH_ERROR("std::make_unique<LoadedModel> error: polyhedra_1.obj");
    }
    auto polyhedra2Model = std::make_unique<LoadedModel>(std::move(texturePhongShader), "polyhedra_2.obj",
                                                         "assets/models");
    if (!polyhedra2Model) {
        EXIT_WITH_ERROR("std::make_unique<LoadedModel> error: polyhedra_2.obj");
    }
    auto polyhedra3Model = std::make_unique<LoadedModel>(std::move(texturePhongShader), "polyhedra_3.obj",
                                                         "assets/models");
    if (!polyhedra3Model) {
        EXIT_WITH_ERROR("std::make_unique<LoadedModel> error: polyhedra_3.obj");
    }
    auto polyhedra4Model = std::make_unique<LoadedModel>(std::move(texturePhongShader), "polyhedra_4.obj",
                                                         "assets/models");
    if (!polyhedra4Model) {
        EXIT_WITH_ERROR("std::make_unique<LoadedModel> error: polyhedra_4.obj");
    }
    auto polyhedra5Model = std::make_unique<LoadedModel>(std::move(texturePhongShader), "polyhedra_5.obj",
                                                         "assets/models");
    if (!polyhedra5Model) {
        EXIT_WITH_ERROR("std::make_unique<LoadedModel> error: polyhedra_5.obj");
    }

    // Pointlights
    auto fireflyLightobjectPointLight = PointLight(glm::vec3(1.0f, 1.0f, 1.0f), glm::vec3(0.0f, -1.0f, -10.68f),
                                                   glm::vec3(1.0f, 0.045f, 0.0075f), false);
    auto mysteryLightcubePointLight = PointLight(glm::vec3(1.0f, 1.0f, 1.0f),
                                                 glm::vec3(-56.942772, -8.862806f, -29.537695),
                                                 glm::vec3(1.0f, 0.09f, 0.032f), true);

    // Pointlights Cubemap Face Projection Transforms
    float depthCubemapAspect = depthCubemapTextureWidth / depthCubemapTextureHeight;
    glm::mat4 depthCubemapProjection = glm::perspective(glm::radians(90.0f), depthCubemapAspect, nearPlane, farPlane);

    // Materials
    PhongIlluminationMaterial caveMaterial{glm::vec3(0.0f, 0.5f, 0.9f), 0.1, 0.7, 0.1, 2};
    PhongIlluminationMaterial lightMaterial{glm::vec3(1.0f, 1.0f, 1.0f), 0.1, 0.7, 0.1, 2};
    caveModel->setMaterial(caveMaterial);
    mysteryLightcubeModel->setMaterial(caveMaterial);
    fireflyLightobjectModel->setMaterial(lightMaterial);

    // Set model positions and keyframe animations:

    // mysteryCubeModel is placed somewhere within the cave and spins around y-axis for 200 seconds
    glm::vec3 lightCubePosition = glm::vec3(-56.942772, -8.862806f, -29.537695);
    mysteryLightcubeModel->setTranslation(lightCubePosition.x, lightCubePosition.y, lightCubePosition.z);
    for (int i = 0; i < camera.keyframes.size() + 3; ++i) {
        float angleRadians = glm::radians(45.0f * i);
        glm::vec3 viewDirection = glm::vec3(sin(angleRadians), sin(angleRadians), -cos(angleRadians));
        mysteryLightcubeModel->addKeyframe(mysteryLightcubeModel->getPosition(), viewDirection, glm::vec3(1.0f), 2.0f);
    }

    fireflyLightobjectModel->setTranslation(0.0f, 3.0f, -10.68f);
    bool enableRandomJitter = true;
    float randomShiftRange = 1.0f;

    // Define a keyframe class
    class PositionedModelKeyframe {
    public:
        glm::vec3 position;
        glm::vec3 viewDirection;
        glm::vec3 scale;
        float transitionTime;

        PositionedModelKeyframe(const glm::vec3 &p, const glm::vec3 &vd, const glm::vec3 &s, const float tt)
            : position(p), viewDirection(glm::normalize(vd)), scale(s), transitionTime(tt) {
        }
    };

    // Array to store raw positions
    std::vector<glm::vec3> positions = {
        glm::vec3(-0.191448, 2.223722, 1.835093),
        glm::vec3(-0.191448, 2.223722, 1.835093),
        glm::vec3(-0.174721, 2.237544, 1.295983),
        glm::vec3(0.136814, 2.388056, -7.698870),
        glm::vec3(0.115963, 1.586614, -17.074049),
        glm::vec3(-0.216977, -0.373172, -24.248470),
        glm::vec3(-3.399364, -1.481182, -29.794794),
        glm::vec3(-10.611661, -1.899825, -31.518883),
        glm::vec3(-19.245300, -2.350358, -31.883879),
        glm::vec3(-29.384750, -3.866306, -31.681662),
        glm::vec3(-38.495831, -5.113851, -31.850531),
        glm::vec3(-47.090527, -5.598403, -32.705860),
        glm::vec3(-53.459328, -5.628543, -33.500725),
        glm::vec3(-56.872849, -5.342404, -29.885204),
        glm::vec3(-56.942772, -4.862806f, -29.537695),
    };

    for (size_t i = 0; i < positions.size(); ++i) {
        glm::vec3 position = positions[i];
        if (enableRandomJitter) {
            glm::vec3 randomShift = glm::linearRand(glm::vec3(-randomShiftRange), glm::vec3(randomShiftRange));
            position += randomShift;
        }
        float angleRadians = glm::radians(45.0f * static_cast<float>(i));
        glm::vec3 viewDirection = glm::vec3(sin(angleRadians), sin(angleRadians), -cos(angleRadians));
        glm::vec3 scale(1.0f, 1.0f, 1.0f);
        fireflyLightobjectModel->addKeyframe(position, viewDirection, scale, 2.0f);
        fireflyLightobjectPointLight.addKeyframe(glm::vec3(position.x, position.y + 1.5f, position.z), 2.0f);
    }

    for (size_t i = positions.size(); i < camera.keyframes.size() + 3; ++i) {
        glm::vec3 position = glm::vec3(-56.942772, -4.862806f, -29.537695);
        float angleRadians = glm::radians(45.0f * static_cast<float>(i));
        glm::vec3 viewDirection = glm::vec3(sin(angleRadians), sin(angleRadians), -cos(angleRadians));
        glm::vec3 randomScale = glm::linearRand(glm::vec3(1), glm::vec3(2));
        fireflyLightobjectModel->addKeyframe(position, viewDirection, randomScale, 2.0f);
        fireflyLightobjectPointLight.addKeyframe(glm::vec3(position.x, position.y + 1.5f, position.z), 2.0f);
    }

    // Arrange polyhedra around the mystery cube
    float radius = 7.0f;
    for (int i = 0; i < 5; ++i) {
        float angleRadians = glm::radians(72.0f * i);
        float xOffset = radius * cos(angleRadians);
        float zOffset = radius * sin(angleRadians);
        glm::vec3 polyhedraPosition = glm::vec3(-56.942772f + xOffset, -8.862806f, -29.537695f + zOffset);
        switch (i) {
            case 0:
                polyhedra1Model->setTranslation(polyhedraPosition.x, polyhedraPosition.y, polyhedraPosition.z);
                polyhedra1Model->setMaterial(caveMaterial);
                break;
            case 1:
                polyhedra2Model->setTranslation(polyhedraPosition.x, polyhedraPosition.y, polyhedraPosition.z);
                polyhedra1Model->setMaterial(caveMaterial);
                break;
            case 2:
                polyhedra3Model->setTranslation(polyhedraPosition.x, polyhedraPosition.y, polyhedraPosition.z);
                polyhedra1Model->setMaterial(caveMaterial);
                break;
            case 3:
                polyhedra4Model->setTranslation(polyhedraPosition.x, polyhedraPosition.y, polyhedraPosition.z);
                polyhedra1Model->setMaterial(caveMaterial);
                break;
            case 4:
                polyhedra5Model->setTranslation(polyhedraPosition.x, polyhedraPosition.y, polyhedraPosition.z);
                polyhedra1Model->setMaterial(caveMaterial);
                break;
        }
    }

    // DrawableScene
    DrawableScene<CatmullRomCamera> scene = DrawableScene{camera};
    scene.addObject(std::move(caveModel));
    scene.addObject(std::move(mysteryLightcubeModel));
    scene.addObject(std::move(fireflyLightobjectModel));
    scene.addObject(std::move(polyhedra1Model));
    scene.addObject(std::move(polyhedra2Model));
    scene.addObject(std::move(polyhedra3Model));
    scene.addObject(std::move(polyhedra4Model));
    scene.addObject(std::move(polyhedra5Model));
    scene.addDirectionalLight({{0.8f, 0.8f, 0.8f}, {-2.0f, 4.0f, -1.0f}});
    scene.addPointLight(mysteryLightcubePointLight);
    scene.addPointLight(fireflyLightobjectPointLight);

    // Keyboard
    bool keys[1024];
    bool useCatmullRomCamera = true;
    std::fill(std::begin(keys), std::end(keys), false);
    auto keyCallback = [&](Window &window, int key, int scancode, int action, int mods) {
        if (key >= 0 && key < 1024) {
            if (action == GLFW_PRESS)
                keys[key] = true;
            else if (action == GLFW_RELEASE)
                keys[key] = false;
        }

        // "Esc": Close window
        if (key == GLFW_KEY_ESCAPE && action == GLFW_PRESS)
            window.setCloseFlag();

        // "C": Toggle between keyframe animated CatmulRomCamera and WASD-camera
        if (key == GLFW_KEY_C && action == GLFW_PRESS)
            useCatmullRomCamera = !useCatmullRomCamera;

        // "P": Print camera position and direction
        if (key == GLFW_KEY_E && action == GLFW_PRESS) {
            glm::vec3 cameraPosition = scene.camera().position;
            glm::vec3 cameraDirection = scene.camera().forward;
            std::cout << "camera.addKeyframe(glm::" << glm::to_string(cameraPosition) << ", glm::" << glm::to_string(
                cameraPosition + cameraDirection) << ", 2.0f); " << std::endl;
        }
    };
    window->setKeyCallback(keyCallback);

    // Mouse
    float lastMouseX = windowWidth / 2.0f;
    float lastMouseY = windowHeight / 2.0f;
    bool mouseInitStatus = false;
    auto cursorCallback = [&](Window &window, double xpos, double ypos) {
        if (mouseInitStatus == false) {
            lastMouseX = float(xpos);
            lastMouseY = float(ypos);
            mouseInitStatus = true;
        }

        float xoffset = float(xpos) - lastMouseX;
        float yoffset = lastMouseY - float(ypos);

        lastMouseX = float(xpos);
        lastMouseY = float(ypos);

        if (!useCatmullRomCamera) {
            scene.camera().processMouseMovement(xoffset, yoffset);
        }
    };
    window->setMouseCallback(cursorCallback);

    // Render loop
    float currentFrameTime = 0.0f;
    float lastFrameTime = 0.0f;
    float fpsTimer = 0.0f;
    int frameCount = 0;

    // Audio
    Audio audio;
    audio.playAudioAsynchronous("assets/audio/mystery.wav");
    audio.setSource(lightCubePosition);
    while (window->keepRunning()) {
        float currentTime = glfwGetTime();
        float deltaTime = currentTime - lastFrameTime;
        lastFrameTime = currentTime;
        glfwPollEvents();

        if (currentTime > camera.keyframes.size() * 2.0f) {
            break;
        }
        frameCount++;
        fpsTimer += deltaTime;

        if (fpsTimer >= 1.0f) {
            float fps = frameCount / fpsTimer;
            std::cout << "FPS: " << fps << std::endl;
            fpsTimer = 0.0f;
            frameCount = 0;
        }

        // Camera movement
        if (!useCatmullRomCamera) {
            scene.camera().processKeyboard(deltaTime, keys);
            glfwSetInputMode(window->handle(), GLFW_CURSOR, GLFW_CURSOR_DISABLED);
        } else {
            currentFrameTime += deltaTime;
            scene.camera().interpolateOrJump(currentFrameTime);
            glfwSetInputMode(window->handle(), GLFW_CURSOR, GLFW_CURSOR_NORMAL);
            mouseInitStatus = true; // Reset mouse when switching back
        }

        audio.setListener(scene.camera().position);

        // Object movement
        scene.interpolateObjectsInScene(currentFrameTime);

        // Light movement
        scene.interpolatePointLights(currentFrameTime);

        // Render scene
        glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

        glPushDebugGroup(GL_DEBUG_SOURCE_APPLICATION, 1, -1, "Shadow Map Pass");
        glViewport(0, 0, depthCubemapTextureWidth, depthCubemapTextureHeight);
        glBindFramebuffer(GL_FRAMEBUFFER, mysteryLightcubeCubemapFBO);
        glClear(GL_DEPTH_BUFFER_BIT);
        scene.drawShadowMap(farPlane, shadowCubemapShader, depthCubemapProjection, 0);
        glPopDebugGroup();

        glPushDebugGroup(GL_DEBUG_SOURCE_APPLICATION, 2, -1, "Shadow Map Pass");
        glViewport(0, 0, depthCubemapTextureWidth, depthCubemapTextureHeight);
        glBindFramebuffer(GL_FRAMEBUFFER, fireflyLightobjectCubemapFBO);
        glClear(GL_DEPTH_BUFFER_BIT);
        scene.drawShadowMap(farPlane, shadowCubemapShader, depthCubemapProjection, 1);
        glPopDebugGroup();

        glPushDebugGroup(GL_DEBUG_SOURCE_APPLICATION, 3, -1, "Color and depth render-pass (sceneFBO)");
        glBindFramebuffer(GL_FRAMEBUFFER, sceneColorDepthFBO);
        glViewport(0, 0, windowWidth, windowHeight);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        scene.draw(farPlane, mysteryLightcubeCubemapTexture, fireflyLightobjectCubemapTexture);
        glPopDebugGroup();

        glPushDebugGroup(GL_DEBUG_SOURCE_APPLICATION, 4, -1, "Shadow Downsampling");
        glBindFramebuffer(GL_FRAMEBUFFER, downsampleDepthFBO);
        glViewport(0, 0, windowWidth / downsamplingFactor, windowHeight / downsamplingFactor);
        glClear(GL_DEPTH_BUFFER_BIT);
        depthDownsampleShader->bind();
        depthDownsampleShader->setSceneDepth(*sceneDepthTexture);
        glBindVertexArray(scene.getFullScreenQuadVAO());
        glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
        glBindVertexArray(0);
        ShaderProgram::unbind();
        glPopDebugGroup();

        glPushDebugGroup(GL_DEBUG_SOURCE_APPLICATION, 5, -1, "God Ray / Post-Processing Pass");
        glBindFramebuffer(GL_FRAMEBUFFER, volumetricGodRayFBO);
        glViewport(0, 0, windowWidth / downsamplingFactor, windowHeight / downsamplingFactor);
        glClear(GL_DEPTH_BUFFER_BIT);
        scene.drawGodRays(farPlane, volumetricGodRayShader, mysteryLightcubeCubemapTexture, downsampledDepthTexture);
        glPopDebugGroup();

        glPushDebugGroup(GL_DEBUG_SOURCE_APPLICATION, 1, -1, "Bilateral Blur Pass");
        glBindFramebuffer(GL_FRAMEBUFFER, bilateralBlurFBO);
        glViewport(0, 0, windowWidth / downsamplingFactor, windowHeight / downsamplingFactor);
        glClear(GL_COLOR_BUFFER_BIT);
        bilateralBlurShader->bind();
        bilateralBlurShader->setInputTexture(*downsampledGodRayTexture);
        bilateralBlurShader->setTexelSize(1.0f / (windowWidth / downsamplingFactor),
                                          1.0f / (windowHeight / downsamplingFactor));
        glBindVertexArray(scene.getFullScreenQuadVAO());
        glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
        glBindVertexArray(0);
        ShaderProgram::unbind();
        glPopDebugGroup();


        glPushDebugGroup(GL_DEBUG_SOURCE_APPLICATION, 6, -1, "Shadow Upsampling");
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
        glViewport(0, 0, windowWidth, windowHeight);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        lightUpsampleShader->bind();
        lightUpsampleShader->setLightTexture(*blurredGodRayTexture);
        lightUpsampleShader->setColorTexture(*sceneColorTexture);
        lightUpsampleShader->setTexelSize(windowWidth / downsamplingFactor, windowHeight / downsamplingFactor);
        glBindVertexArray(scene.getFullScreenQuadVAO());
        glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
        glBindVertexArray(0);
        ShaderProgram::unbind();
        glPopDebugGroup();

        window->swapBuffers();
    }

    audio.gracefullyStopAudio();
    destroyFramework();
    window = nullptr;
    glfwTerminate();
    return EXIT_SUCCESS;
}
