#pragma once
#ifndef POINTLIGHT_H
#define POINTLIGHT_H

#include <glm/ext.hpp>
#include <glm/gtx/string_cast.hpp>
#include <vector>

class PointLightKeyframe {
public:
	glm::vec3 position;
	float transitionTime;

	PointLightKeyframe(const glm::vec3& p, const float tt)
		: position(p), transitionTime(tt) {}

};

class PointLight {
public:
    glm::vec3 intensity;
    glm::vec3 position;
    glm::vec3 attenuationFactors;
    bool volumetricGodRay;
    std::vector<PointLightKeyframe> keyframes;
	PointLight(const glm::vec3& i, const glm::vec3& p, const glm::vec3& a, bool volumetricGodRay)
		: intensity(i), position(p), attenuationFactors(a), volumetricGodRay(volumetricGodRay) {}

    void interpolateOrJump(float currentTime) {
        if (keyframes.empty()) return;

        int k1, k2;
        float localTime;
        findKeyframes(currentTime, keyframes, k1, k2, localTime);


        if (keyframes[k2].transitionTime == 0) {
            glm::vec3 newPosition = keyframes[k2].position;
            updatePointLight(newPosition);
        }
        else {
            interpolatePointLight(currentTime, keyframes);
        }
    }

    void addKeyframe(const glm::vec3& position, const float transitionTime) {
        keyframes.emplace_back(PointLightKeyframe(position, transitionTime));
    }

protected:
    glm::vec3 catmullRomSpline(const glm::vec3& P0, const glm::vec3& P1, const glm::vec3& P2, const glm::vec3& P3, float t) {
        return 0.5f * ((2.0f * P1) + (-P0 + P2) * t + (2.0f * P0 - 5.0f * P1 + 4.0f * P2 - P3) * (t * t) + (-P0 + 3.0f * P1 - 3.0f * P2 + P3) * (t * t * t));
    }

    glm::quat slerpOrientation(const glm::quat& q0, const glm::quat& q1, float t) {
        return glm::slerp(q0, q1, t);
    }

    void findKeyframes(float currentTime, const std::vector<PointLightKeyframe>& keyframes, int& k1, int& k2, float& localTime) {
        float accumulatedTime = 0.0f;
 
        for (int i = 0; i < keyframes.size() - 3; i++) {
            accumulatedTime += keyframes[i].transitionTime;

            if (currentTime < accumulatedTime) {
                k1 = i + 1;
                k2 = i + 2;
                localTime = (currentTime - (accumulatedTime - keyframes[i + 1].transitionTime)) / keyframes[i + 1].transitionTime;
                return;
            }
        }

        k1 = keyframes.size() - 3;
        k2 = keyframes.size() - 2;
        localTime = 1.0f;
    }

    void interpolatePointLight(float currentTime, const std::vector<PointLightKeyframe>& keyframes) {
        int k1, k2;
        float localTime;
        findKeyframes(currentTime, keyframes, k1, k2, localTime);

        PointLightKeyframe kf0 = keyframes[k1 - 1];
        PointLightKeyframe kf1 = keyframes[k1];
        PointLightKeyframe kf2 = keyframes[k2];
        PointLightKeyframe kf3 = keyframes[k2 + 1];

        glm::vec3 newPosition = catmullRomSpline(kf0.position, kf1.position, kf2.position, kf3.position, localTime);

        updatePointLight(newPosition);
    }


    void updatePointLight(const glm::vec3& newPosition) {
        position = newPosition;
    }

private:
};


#endif