#pragma once
#ifndef POSITONEDMODEL_H
#define POSITONEDMODEL_H

#include <glm/ext.hpp>
#include <glm/gtx/string_cast.hpp>
#include <vector>
#include <algorithm>

class PositionedModelKeyframe {
public:
    glm::vec3 position;
    glm::vec3 viewDirection; // Direction the object should face
    glm::vec3 scale;
    float transitionTime; // Time to transition to the next keyframe

    PositionedModelKeyframe(const glm::vec3& p, const glm::vec3& vd, const glm::vec3& s, const float tt)
        : position(p), viewDirection(glm::normalize(vd)), scale(s), transitionTime(tt) {}
};

class LoadedModel;

class PositionedModel {
public:
    PositionedModel(const glm::vec3& s, const glm::vec3& vd, const glm::vec3& p)
        : scaling(s), viewDirection(glm::normalize(vd)), position(p) {
        calcMatrix();
    }

    PositionedModel()
        : scaling(glm::vec3{ 1.0f, 1.0f, 1.0f }),
        viewDirection(glm::vec3{ 0.0f, 0.0f, -1.0f }),
        position(glm::vec3{ 0.0f, 0.0f, 0.0f }) {
        calcMatrix();
    }

    const glm::mat4& modelMatrix() const {
        calcMatrixIfNeeded();
        return matrix;
    }

    PositionedModel& setTranslation(float x, float y, float z) {
        position = glm::vec3{ x, y, z };
        matrixNeedsCalc = true;
        return *this;
    }

    PositionedModel& setViewDirection(const glm::vec3& vd) {
        viewDirection = glm::normalize(vd);
        matrixNeedsCalc = true;
        return *this;
    }

    PositionedModel& scale(float x, float y, float z) {
        scaling = glm::vec3{ x, y, z };
        matrixNeedsCalc = true;
        return *this;
    }

    void findKeyframes(float currentTime, const std::vector<PositionedModelKeyframe>& keyframes, int& k1, int& k2, float& localTime) {
        float accumulatedTime = 0.0f;

        for (int i = 0; i < keyframes.size() - 3; i++) {
            accumulatedTime += keyframes[i].transitionTime;

            if (currentTime < accumulatedTime) {
                k1 = i + 1;
                k2 = i + 2;
                localTime = (currentTime - (accumulatedTime - keyframes[i + 1].transitionTime)) / keyframes[i + 1].transitionTime;
                return;
            }
        }

        k1 = keyframes.size() - 3;
        k2 = keyframes.size() - 2;
        localTime = 1.0f;
        return;
    }

    float totalKeyframeTime() const {
        float total = 0.0f;
        for (const auto& kf : keyframes)
            total += kf.transitionTime;
        return total;
    }

    void interpolateOrJump(float currentTime) {
        if (keyframes.empty()) return;

        if (currentTime >= totalKeyframeTime()) {
            const PositionedModelKeyframe& lastKeyframe = keyframes.back();
            glm::vec3 newPosition = lastKeyframe.position;
            glm::vec3 newViewDir = lastKeyframe.viewDirection;
            glm::vec3 newScale = lastKeyframe.scale;
            updateObject(newPosition, newScale, newViewDir);
        }
        else {
            int k1, k2;
            float localTime;
            findKeyframes(currentTime, keyframes, k1, k2, localTime);

            if (keyframes[k2].transitionTime == 0) {
                glm::vec3 newPosition = keyframes[k2].position;
                glm::vec3 newViewDir = keyframes[k2].viewDirection;
                glm::vec3 newScale = keyframes[k2].scale;
                updateObject(newPosition, newScale, newViewDir);
            }
            else {
                interpolateObject(currentTime, keyframes);
            }
        }
    }

    glm::vec3 catmullRomSpline(const glm::vec3& P0, const glm::vec3& P1, const glm::vec3& P2, const glm::vec3& P3, float t) {
        return 0.5f * ((2.0f * P1) + (-P0 + P2) * t + (2.0f * P0 - 5.0f * P1 + 4.0f * P2 - P3) * (t * t) + (-P0 + 3.0f * P1 - 3.0f * P2 + P3) * (t * t * t));
    }

    void interpolateObject(float currentTime, const std::vector<PositionedModelKeyframe>& keyframes) {
        int k1, k2;
        float localTime;
        findKeyframes(currentTime, keyframes, k1, k2, localTime);

        PositionedModelKeyframe kf0 = keyframes[k1 - 1];
        PositionedModelKeyframe kf1 = keyframes[k1];
        PositionedModelKeyframe kf2 = keyframes[k2];
        PositionedModelKeyframe kf3 = keyframes[k2 + 1];

        glm::vec3 newPosition = catmullRomSpline(kf0.position, kf1.position, kf2.position, kf3.position, localTime);
        glm::vec3 newScale = catmullRomSpline(kf0.scale, kf1.scale, kf2.scale, kf3.scale, localTime);
        glm::vec3 interpolatedViewDir = catmullRomSpline(kf0.viewDirection, kf1.viewDirection, kf2.viewDirection, kf3.viewDirection, localTime);

        interpolatedViewDir = glm::normalize(interpolatedViewDir);
        updateObject(newPosition, newScale, interpolatedViewDir);
    }

    void addKeyframe(const glm::vec3& position, const glm::vec3& viewDirection, const glm::vec3& scale, const float transitionTime) {
        keyframes.emplace_back(PositionedModelKeyframe(position, viewDirection, scale, transitionTime));
    }

    PositionedModel& updateObject(const glm::vec3& newPosition, const glm::vec3& newScale, const glm::vec3& newViewDir) {
        setTranslation(newPosition.x, newPosition.y, newPosition.z);
        setViewDirection(newViewDir);
        scale(newScale.x, newScale.y, newScale.z);

        return *this;
    }

    glm::vec3 getPosition() const {
        return position;
    }

protected:
    std::vector<PositionedModelKeyframe> keyframes;

    void calcMatrix() const {
        matrix = glm::mat4{ 1.0f };
        matrix = glm::translate(matrix, position);

        glm::vec3 up = glm::vec3{ 0.0f, 1.0f, 0.0f };
        glm::vec3 zAxis = -glm::normalize(viewDirection);
        glm::vec3 xAxis = glm::normalize(glm::cross(up, zAxis));
        glm::vec3 yAxis = glm::cross(zAxis, xAxis);

        glm::mat4 rotationMatrix = glm::mat4(
            glm::vec4(xAxis, 0.0f),
            glm::vec4(yAxis, 0.0f),
            glm::vec4(zAxis, 0.0f),
            glm::vec4(0.0f, 0.0f, 0.0f, 1.0f)
        );

        matrix *= rotationMatrix;
        matrix = glm::scale(matrix, scaling);
    }

    void calcMatrixIfNeeded() const {
        if (matrixNeedsCalc) {
            calcMatrix();
            matrixNeedsCalc = false;
        }
    }

private:
    glm::vec3 scaling;
    glm::vec3 viewDirection;
    glm::vec3 position;
    mutable bool matrixNeedsCalc = false;
    mutable glm::mat4 matrix;
};

#endif