#pragma once
#ifndef SCENE_H
#define SCENE_H

#include "TextureSceneColor.h"
#include "TextureSceneDepth.h"
#include "ShaderVolumetricGodRay.h"
#include "ShaderShadowCubemap.h"

#endif
template<typename T>
class DrawableScene {
public:
	DrawableScene(T cam)
		: sceneCamera{ cam } {
			setupFullScreenQuad();
		}

	void addObject(std::unique_ptr<Drawable> obj) {
		objects.emplace_back(std::move(obj));
	}

	void addDirectionalLight(DirectionalLight dl) {
		directionalLights.emplace_back(dl);
	}

	void addPointLight(PointLight &pl) {
		pointLights.emplace_back(std::move(pl));
	}

	void draw(float far, std::shared_ptr<TextureShadowCubemap> mysteryLightcubeCubemapTexture, std::shared_ptr<TextureShadowCubemap> fireflyLightobjectCubemapTexture) {
		for (auto& obj : objects) {
			obj->draw(sceneCamera, far, directionalLights, pointLights, mysteryLightcubeCubemapTexture, pointLights[0].position, fireflyLightobjectCubemapTexture, pointLights[1].position);
		}
	}

	void drawGodRays(float far, const std::shared_ptr<ShaderVolumetricGodray>& shader, const std::shared_ptr<TextureShadowCubemap>& cubeMapTexture, const std::shared_ptr<TextureSceneDepth>& downsampledDepthTexture) {
		glm::mat4 lightSpaceMatrix = glm::mat4(1.0); //identity matrix
		shader->bind();
		shader->setDownsampledDepth(*downsampledDepthTexture);
		shader->setShadowCubemap(*cubeMapTexture);
		shader->setLightSpaceMatrix(lightSpaceMatrix);
		shader->setFarPlane(far);
		shader->setLightPosition(pointLights[0].position);
		glm::mat4 viewProj = sceneCamera.projectionMatrix * sceneCamera.viewMatrix;
		glm::mat4 inverseVP = glm::inverse(viewProj);
		shader->setInverseViewProjection(inverseVP);
		shader->setCameraPosition(sceneCamera.position);
		glBindVertexArray(quadVAO);
		glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
		glBindVertexArray(0);
		ShaderProgram::unbind();
	}

    void drawShadowMap(float far, const std::shared_ptr<ShaderShadowCubemap>& shader, glm::mat4 depthCubemapProjection, int pointLightIndex) {
        for (auto& obj : objects) {
			glm::vec3 pointLightPosition = pointLights[pointLightIndex].position;
			std::vector<glm::mat4> cubefaceTransforms;
			cubefaceTransforms.push_back(depthCubemapProjection * glm::lookAt(pointLightPosition, pointLightPosition + glm::vec3(1.0f, 0.0f, 0.0f), glm::vec3(0.0f, -1.0f, 0.0f)));
			cubefaceTransforms.push_back(depthCubemapProjection * glm::lookAt(pointLightPosition, pointLightPosition + glm::vec3(-1.0f, 0.0f, 0.0f), glm::vec3(0.0f, -1.0f, 0.0f)));
			cubefaceTransforms.push_back(depthCubemapProjection * glm::lookAt(pointLightPosition, pointLightPosition + glm::vec3(0.0f, 1.0f, 0.0f), glm::vec3(0.0f, 0.0f, 1.0f)));
			cubefaceTransforms.push_back(depthCubemapProjection * glm::lookAt(pointLightPosition, pointLightPosition + glm::vec3(0.0f, -1.0f, 0.0f), glm::vec3(0.0f, 0.0f, -1.0f)));
			cubefaceTransforms.push_back(depthCubemapProjection * glm::lookAt(pointLightPosition, pointLightPosition + glm::vec3(0.0f, 0.0f, 1.0f), glm::vec3(0.0f, -1.0f, 0.0f)));
			cubefaceTransforms.push_back(depthCubemapProjection * glm::lookAt(pointLightPosition, pointLightPosition + glm::vec3(0.0f, 0.0f, -1.0f), glm::vec3(0.0f, -1.0f, 0.0f)));

        	shader->bind();
            obj->drawInLightPerspective(far, shader, pointLightPosition, cubefaceTransforms);
        	ShaderProgram::unbind();
        }
    }

	T& camera() { return sceneCamera; }

    void interpolateObjectsInScene(double t) {

        for (auto& obj : objects) {
            obj->interpolateOrJump(t);
        }
    }

    void interpolatePointLights(double t) {
        for (auto& light : pointLights) {
            light.interpolateOrJump(t);
        }
    }
	
	void setupFullScreenQuad() {
		float quadVertices[] = {
			// Positions    // Texture Coordinates
			-1.0f,  1.0f,   0.0f, 1.0f, // Top-left
			-1.0f, -1.0f,   0.0f, 0.0f, // Bottom-left
			1.0f, -1.0f,   1.0f, 0.0f, // Bottom-right
			1.0f,  1.0f,   1.0f, 1.0f  // Top-right
		};

		glGenVertexArrays(1, &quadVAO);
		glGenBuffers(1, &quadVBO);

		glBindVertexArray(quadVAO);
		glBindBuffer(GL_ARRAY_BUFFER, quadVBO);
		glBufferData(GL_ARRAY_BUFFER, sizeof(quadVertices), &quadVertices, GL_STATIC_DRAW);

		glEnableVertexAttribArray(0);
		glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 4 * sizeof(float), (void*)0);

		glEnableVertexAttribArray(1);
		glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 4 * sizeof(float), (void*)(2 * sizeof(float)));

		glBindVertexArray(0);
	}

	GLuint getFullScreenQuadVAO() const {
		return quadVAO;
	}

private:
	T sceneCamera;
	std::vector<std::unique_ptr<Drawable>> objects;
	std::vector<DirectionalLight> directionalLights;
	std::vector<PointLight> pointLights;
	GLuint quadVAO, quadVBO;
};