#pragma once
#ifndef SHADER_H
#define SHADER_H

#include "Utils.h"
#include "Bindable.h"
#include <glm/ext.hpp>
#include <glm/gtx/string_cast.hpp>
#include <sstream>
#include <optional>
#include <fstream>

std::string readShaderFile( const std::string& path ) {
    std::ifstream file(path);
    if (!file.is_open()) {
        throw std::runtime_error("Could not open shader source file with path" + path);
    }

    auto stringReader = std::ostringstream{};
    stringReader << file.rdbuf();
    return stringReader.str();
}

class ShaderAttribute {
public:
	ShaderAttribute(GLint _id) : id(_id) {}

	GLint handle() const { return id; }
	operator bool() const { return id >= 0; }
protected:
	GLint id;
};

class ShaderUniformBase {
public:
	ShaderUniformBase(GLint _id = -1) : id(_id) {}

	GLint handle() const { return id; }
	operator bool() const { return id >= 0; }
protected:
	GLint id;
};

template<typename T>
class ShaderUniform : public ShaderUniformBase {};

template<>
class ShaderUniform<glm::vec3> : public ShaderUniformBase {
public:
	using ShaderUniformBase::ShaderUniformBase;

	void set(const glm::vec3& data) {
		glUniform3fv(id, 1, glm::value_ptr(data));
	}
};

template<>
class ShaderUniform<glm::vec2> : public ShaderUniformBase {
public:
	using ShaderUniformBase::ShaderUniformBase;

	void set(const glm::vec2& data) {
		glUniform2fv(id, 1, glm::value_ptr(data));
	}
};


template<>
class ShaderUniform<glm::mat4> : public ShaderUniformBase {
public:
	using ShaderUniformBase::ShaderUniformBase;

	void set(const glm::mat4& data) {
		glUniformMatrix4fv(id, 1, GL_FALSE, glm::value_ptr(data));
	}
};

template<>
class ShaderUniform<float> : public ShaderUniformBase {
public:
	using ShaderUniformBase::ShaderUniformBase;

	void set(float data) {
		glUniform1f(id, data);
	}
};

template<>
class ShaderUniform<uint32_t> : public ShaderUniformBase {
public:
	using ShaderUniformBase::ShaderUniformBase;

	void set(uint32_t data) {
		glUniform1ui(id, data);
	}
};

template<>
class ShaderUniform<int32_t> : public ShaderUniformBase {
public:
	using ShaderUniformBase::ShaderUniformBase;

	void set(int32_t data) {
		glUniform1i(id, data);
	}
};

template<>
class ShaderUniform<bool> : public ShaderUniformBase {
public:
	using ShaderUniformBase::ShaderUniformBase;

	void set(bool data) {
		glUniform1i(id, data);
	}
};

class Shader {
public:
	static std::optional<Shader> createFromSource(GLenum shaderType, const std::string& source) {
		const auto handle = glCreateShader(shaderType);
		if (!handle) {
			return {};
		}

		const auto* sourceString = source.c_str();
		glShaderSource(handle, 1, &sourceString, NULL);
		glCompileShader(handle);

		GLint succeded;
		glGetShaderiv(handle, GL_COMPILE_STATUS, &succeded);
		if (succeded == GL_FALSE) {
			GLint logSize;
			glGetShaderiv(handle, GL_INFO_LOG_LENGTH, &logSize);

			std::string message;
			message.resize(logSize);

			glGetShaderInfoLog(handle, logSize, NULL, const_cast<char*>(message.c_str()));
			std::cerr << "Shader compile error: " << message << std::endl;
			return {};
		}

		return { Shader(handle) };
	}

	Shader(const Shader&) = delete;
	Shader(Shader&& s) noexcept {
		shader = s.shader;
		s.shader = -1;
	}

	Shader(GLuint handle)
		: shader(handle) {}

	virtual ~Shader() {
		if (shader != -1) {
			glDeleteShader(shader);
		}
	}

private:
	friend class ShaderProgram;

	GLuint shader;
};

class ShaderProgram : public Bindable {
public:

	template<typename T = ShaderProgram>
	static std::unique_ptr<T> createFromFileName(std::string fileName, bool hasGeomShader) {
		fileName += ".vert";
		auto vertexShader = Shader::createFromSource(GL_VERTEX_SHADER, readShaderFile(fileName));
		if (!vertexShader.has_value()) {
			return { nullptr };
		}

		fileName.resize(fileName.size() - 5);
		fileName += ".frag";
		auto fragmentShader = Shader::createFromSource(GL_FRAGMENT_SHADER, readShaderFile(fileName));
		if (!fragmentShader.has_value()) {
			return { nullptr };
		}

		fileName.resize(fileName.size() - 5);
		fileName += ".geom";
		auto geomShader = hasGeomShader ? Shader::createFromSource(GL_GEOMETRY_SHADER, readShaderFile(fileName)) : std::nullopt;
		if (hasGeomShader && !geomShader.has_value()) {
			return { nullptr };
		}


		GLuint handle = glCreateProgram();
		glAttachShader(handle, vertexShader.value().shader);
		glAttachShader(handle, fragmentShader.value().shader);
		if (hasGeomShader) {
			glAttachShader(handle, geomShader.value().shader);
		}
		glLinkProgram(handle);

		GLint succeded;
		glGetProgramiv(handle, GL_LINK_STATUS, &succeded);
		if (!succeded) {
			GLint logSize;
			glGetProgramiv(handle, GL_INFO_LOG_LENGTH, &logSize);

			std::string message;
			message.resize(logSize);

			glGetProgramInfoLog(handle, logSize, NULL, const_cast<char*>(message.c_str()));
			std::cerr << "Shader link error: " << message << std::endl;
			return { nullptr };
		}


		glDetachShader(handle, vertexShader.value().shader);
		glDetachShader(handle, fragmentShader.value().shader);
		if (hasGeomShader) {
			glDetachShader(handle, geomShader.value().shader);
		}
		if (geomShader) {
			return std::unique_ptr<T>(new T(handle, std::move(vertexShader.value()), std::move(fragmentShader.value()), std::move(geomShader.value())));

		}

		return std::unique_ptr<T>(new T(handle, std::move(vertexShader.value()), std::move(fragmentShader.value())));

	}

	virtual ~ShaderProgram() {
		glDeleteProgram(shaderProgram);
	}

	virtual void bind() const override {
		glUseProgram(shaderProgram);
	}

	static void unbind() {
		glUseProgram(0);
	}

	template<typename T>
	ShaderUniform<T> uniformLocation(const char* name, GLint defaultLocation = -1) const {
		auto id = glGetUniformLocation(shaderProgram, name);
		return ShaderUniform<T>{ id >= 0 ? id : defaultLocation };
	}

	ShaderAttribute attributeLocation(const char* name, GLint defaultLocation = -1) const {
		auto id = glGetAttribLocation(shaderProgram, name);
		return { id >= 0 ? id : defaultLocation };
	}

	ShaderProgram(const ShaderProgram&) = delete;

protected:
	ShaderProgram(GLuint handle, Shader vert, Shader frag, Shader geomShader)
		: shaderProgram(handle), vertexShader(std::move(vert)), fragmentShader(std::move(frag)), geomShader(std::move(geomShader)) {}

	ShaderProgram(GLuint handle, Shader vert, Shader frag)
		: shaderProgram(handle), vertexShader(std::move(vert)), fragmentShader(std::move(frag)) {}

private:
	GLuint shaderProgram;
	Shader vertexShader;
	Shader fragmentShader;
	std::optional<Shader> geomShader;
};

template<int TCount, typename ...TItems>
class ShaderUniformStructArray {
	static constexpr int TTupleSize = sizeof...(TItems);

public:
	void init(const ShaderProgram& shader, const char* arrayName, const std::array<const char*, TTupleSize>& names) {

		std::string buffer = arrayName;
		buffer += '[';

		auto basePartIdx = buffer.length();
		for (int i = 0; i != TCount; i++) {
			buffer.resize(basePartIdx);
			buffer += std::to_string(i) + "].";
			auto namePartIdx = buffer.length();


			initUniformTuple<0, TItems...>(shader, i, names, buffer, namePartIdx);
		}
	}

	template<int FieldIdx>
	auto& getFieldFromSlot(int arrayIdx) {
		static_assert(FieldIdx < TTupleSize, "UniformStructArray field access out of bounds");
		return std::get<FieldIdx>(uniforms.at(arrayIdx));
	}

private:
	template<int TupleIdx, typename T>
	void initUniform(const ShaderProgram& shader, int arrayIndex, const std::array<const char*, TTupleSize>& names, std::string& buffer, int namePartIdx) {
		buffer.resize(namePartIdx);
		buffer += names[TupleIdx];
		auto uniformObj = shader.uniformLocation<T>(buffer.c_str());
		getFieldFromSlot<TupleIdx>(arrayIndex) = uniformObj;

		if (!uniformObj) {
			std::cerr << "Invalid uniform for struct array item: " << buffer << std::endl;
		}
	}

	template<int TupleIdx, typename T, typename ...Ts>
	void initUniformTuple(const ShaderProgram& shader, int arrayIndex, const std::array<const char*, TTupleSize>& names, std::string& buffer, int namePartIdx) {
		initUniform<TupleIdx, T>(shader, arrayIndex, names, buffer, namePartIdx);

		if constexpr (TupleIdx < TTupleSize - 1) {
			initUniformTuple<TupleIdx + 1, Ts...>(shader, arrayIndex, names, buffer, namePartIdx);
		}
	}

	std::array<std::tuple< ShaderUniform<TItems>...>, TCount> uniforms;
};


#endif
