#pragma once
#ifndef TEXTURE_ILLUMINATION_SHADER_H
#define TEXTURE_ILLUMINATION_SHADER_H

#include "Utils.h"
#include "Shader.h"
#include "Texture.h"
#include "Materials.h"
#include "Camera.h"
#include "TextureShadowCubemap.h"
#include <array>


class ShaderTexturePhong final : public ShaderProgram {
public:
	static constexpr int maxDirectionalLightCount = 6;
	static constexpr int maxPointLightCount = 6;

	virtual ~ShaderTexturePhong() {}


	void setMysteryLightcubeCubemap(const TextureShadowCubemap& cubemapTexture) {
		glActiveTexture(GL_TEXTURE0 + 3);
		cubemapTexture.bindToUnit(mysteryLightcubeCubemapUniform, 3);
	}

	void setFireflyLightobjectCubemap(const TextureShadowCubemap& cubemapTexture) {
		glActiveTexture(GL_TEXTURE0 + 4);
		cubemapTexture.bindToUnit(fireflyLightobjectCubemapUniform, 4);
	}

	void setAmbientStrength(float strength) {
		ambientStrengthUniform.set(strength);
	}

	void setTexture(const Texture& diffuse, const Texture& specular, const float shininess) {
		glActiveTexture(GL_TEXTURE0 + 5);
		diffuse.bindToUnit(diffuseUniform, 5);
		glActiveTexture(GL_TEXTURE0 + 6);
		specular.bindToUnit(specularUniform, 6);
		shininessUniform.set(shininess);
	}

	void setNormalMap(const Texture& normal) {
		glActiveTexture(GL_TEXTURE0 + 7);
		normal.bindToUnit(normalUniform, 7);
	}

	void setHasNormalMap(bool hasNormalMap) {
		this->hasNormalMap.set(hasNormalMap);
	}

	void addTexture(const Texture& diffuse, const Texture& specular, const float shininess, int index) {
		setTexture(diffuse, specular, shininess);
	}

    void setLightSpaceMatrix(const glm::mat4 matrix) {
        lightSpaceMatrixUniform.set(matrix);
    }


	void setViewPosition(const glm::vec3& position) {
		viewPositionUniform.set(position);
	}

	void setMysteryLightcubePointLightPosition(const glm::vec3& position) {
		mysteryLightcubePointLightPositionUniform.set(position);
	}

	void setFireflyLightobjectPointLightPosition(const glm::vec3& position) {
		fireflyLightobjectPointLightPositionUniform.set(position);
	}

	void setTransformationMatrices(const CatmullRomCamera::TransformMatrices& transformMatrices) {
		projectionViewMatUniform.set(transformMatrices.projectionViewMatrix);
		modelMatUniform.set(transformMatrices.modelMatrix);
	}


	void setDirectionalLights(const std::vector<DirectionalLight>& directionalLights) {
		if (directionalLights.size() > maxDirectionalLightCount) {
			throw std::runtime_error("Too many directional lights for simple illumination phongShader");
		}

		directionalLightCountUniform.set(directionalLights.size());

		for (int i = 0; i != directionalLights.size(); i++) {
			directionalLightsArray.getFieldFromSlot<0>(i).set(directionalLights[i].direction);
			directionalLightsArray.getFieldFromSlot<1>(i).set(directionalLights[i].intensity);
		}
	}

	void setPointLights(const std::vector<PointLight>& pointLights) {
		if (pointLights.size() > maxPointLightCount) {
			throw std::runtime_error("Too many point lights for simple illumination phongShader");
		}

		pointLightCountUniform.set(pointLights.size());

		for (int i = 0; i != pointLights.size(); i++) {

			pointLightsArray.getFieldFromSlot<0>(i).set(pointLights[i].position);
			pointLightsArray.getFieldFromSlot<1>(i).set(pointLights[i].intensity);
			pointLightsArray.getFieldFromSlot<2>(i).set(pointLights[i].attenuationFactors);
		}
	}

	void setFarPlane(float far) {
		farPlane.set(far);
	}

private:
	friend class ShaderProgram;

	ShaderTexturePhong(GLuint handle, Shader vert, Shader frag)
		: ShaderProgram{ handle, std::move(vert), std::move(frag) } {
		initUniforms();
	}

	ShaderTexturePhong(GLuint handle, Shader vert, Shader frag, Shader geom)
		: ShaderProgram{ handle, std::move(vert), std::move(frag), std::move(geom) } {
		initUniforms();
	}

	void initUniforms() {
		mysteryLightcubeCubemapUniform = uniformLocation<int32_t>("mysteryLightcubeCubemap");
		fireflyLightobjectCubemapUniform = uniformLocation<int32_t>("fireflyLightobjectCubemap");
		mysteryLightcubePointLightPositionUniform = uniformLocation<glm::vec3>("mysteryLightcubePointLightPosition");
		fireflyLightobjectPointLightPositionUniform = uniformLocation<glm::vec3>("fireflyLightobjectPointLightPosition");

        lightSpaceMatrixUniform = uniformLocation<glm::mat4>("lightSpaceMatrix");
		viewPositionUniform = uniformLocation<glm::vec3>("viewPosition");
		ambientStrengthUniform = uniformLocation<float>("ambientStrength");
		projectionViewMatUniform = uniformLocation<glm::mat4>("projectionViewMat");
		modelMatUniform = uniformLocation<glm::mat4>("modelMat");

		farPlane = uniformLocation<float>("far_plane");

		diffuseUniform = uniformLocation<int32_t>("material.diffuse");
		specularUniform = uniformLocation<int32_t>("material.specular");
		shininessUniform = uniformLocation<float>("material.shininess");
		normalUniform = uniformLocation<int32_t>("material.normal");
		hasNormalMap = uniformLocation<bool>("material.hasNormalMap");

		directionalLightCountUniform = uniformLocation<uint32_t>("directionalLightCount");
		directionalLightsArray.init(*this, "directionalLights", {"direction", "intensity" });

		pointLightCountUniform = uniformLocation<uint32_t>("pointLightCount");
		pointLightsArray.init(*this, "pointLights", {"position", "intensity", "attenuationFactors" });
	}

	ShaderUniform<int32_t> mysteryLightcubeCubemapUniform;
	ShaderUniform<int32_t> fireflyLightobjectCubemapUniform;

	ShaderUniform<glm::vec3> mysteryLightcubePointLightPositionUniform;
	ShaderUniform<glm::vec3> fireflyLightobjectPointLightPositionUniform;

    ShaderUniform<glm::mat4> lightSpaceMatrixUniform;
	ShaderUniform<glm::vec3> viewPositionUniform;
	ShaderUniform<float> ambientStrengthUniform;
	ShaderUniform<glm::mat4> projectionViewMatUniform;
	ShaderUniform<glm::mat4> modelMatUniform;

	ShaderUniform<float> farPlane;

	ShaderUniform<int32_t> diffuseUniform;
	ShaderUniform<int32_t> specularUniform;
	ShaderUniform<int32_t> normalUniform;
	ShaderUniform<float> shininessUniform;

	ShaderUniform<bool> hasNormalMap;

	ShaderUniform<uint32_t> directionalLightCountUniform;
	ShaderUniformStructArray< maxDirectionalLightCount, glm::vec3, glm::vec3 > directionalLightsArray;

	ShaderUniform<uint32_t> pointLightCountUniform;
	ShaderUniformStructArray< maxPointLightCount, glm::vec3, glm::vec3, glm::vec3 > pointLightsArray;
};
#endif