#pragma once
#ifndef TEXTURE_H
#define TEXTURE_H

#include "Bindable.h"
#include "Utils.h"
#include "Shader.h"

#define STB_IMAGE_IMPLEMENTATION
#include "stb_image.h"

class Texture : public Bindable {
public:

	std::string type;
	std::string path;

	static std::string getFileType(const char *path) {
		std::string strPath(path);
		size_t dotPosition = strPath.find_last_of(".");
		std::string extension = strPath.substr(dotPosition + 1);
		for (char &c : extension) {
			c = std::tolower(c);
		}
		return extension;
	}

	static std::shared_ptr<Texture> createAndBindDDSTexture(GLuint handle, const char * path) {
		std::shared_ptr<Texture> texture{ new Texture(handle) };
		texture->path = path;

		auto image = loadDDS(path);

		if (!image.data || !image.size) {
			std::cerr << "Could not load dds image for texture" << std::endl;
			return {};
		}

		texture->bind();
		glCompressedTexImage2D(
			GL_TEXTURE_2D, 0, GL_COMPRESSED_RGBA_S3TC_DXT1_EXT,
			image.width, image.height, 0, image.size, image.data
		);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);

		glGenerateMipmap(GL_TEXTURE_2D);

		unbind();
		return std::move(texture);
	}

	static std::shared_ptr<Texture> createAndBindTexture(GLuint handle, const char * path) {

		std::shared_ptr<Texture> texture{ new Texture(handle) };
		texture->path = path;

		int width, height, nrComponents;
		unsigned char *data = stbi_load(path, &width, &height, &nrComponents, 0);
		if (!data) {
			std::cout << "Texture failed to load at path: " << path << std::endl;
			path = "assets/textures/default.jpg";
			data = stbi_load(path, &width, &height, &nrComponents, 0);
		}
		if (data)
		{
			GLenum format;
			if (nrComponents == 1)
				format = GL_RED;
			else if (nrComponents == 3)
				format = GL_RGB;
			else if (nrComponents == 4)
				format = GL_RGBA;

			texture->bind();
			glTexImage2D(GL_TEXTURE_2D, 0, format, width, height, 0, format, GL_UNSIGNED_BYTE, data);
			glGenerateMipmap(GL_TEXTURE_2D);

			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

			stbi_image_free(data);
		}
		else
		{
			std::cout << "Texture failed to load" << std::endl;
			stbi_image_free(data);
		}

		return std::move(texture);
	}

	static std::shared_ptr<Texture> create(const char* path) {
		GLuint handle;
		glGenTextures(1, &handle);
		if (!handle) {
			std::cerr << "Could not create gl texture" << std::endl;
			return {};
		}
		auto extension = getFileType(path);
		if (extension == "dds") {
			auto texture = createAndBindDDSTexture(handle, path);
			return std::move(texture);
		} else {
			auto texture = createAndBindTexture(handle, path);
			return std::move(texture);
		}
	}

	virtual ~Texture() {
		if (textureHandle != -1) {
			glDeleteTextures(1, &textureHandle);
		}
	}

	virtual void bind() const override {
		glBindTexture(GL_TEXTURE_2D, textureHandle);
	}

	static void unbind() {
		glBindTexture(GL_TEXTURE_2D, 0);
	}

	void bindToUnit(ShaderUniform<int32_t>& uniform, int unit) const {
		uniform.set(unit);
		glActiveTexture(GL_TEXTURE0 + unit);
		bind();
	}


private:
	Texture(GLuint handle) : textureHandle(handle) {}
	Texture(const Texture&) = delete;
	Texture(Texture&& t) : textureHandle(t.textureHandle) { textureHandle = -1; }

	GLuint textureHandle;
};
#endif