#pragma once
#ifndef WINDOW_H
#define WINDOW_H
#include <sstream>
#include <functional>
#include "Utils.h"
#include "Bindable.h"

class Window final : public Bindable {
public:
	static std::unique_ptr<Window> create(int width, int height, std::string title) {
		glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 4);
		glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 5);
		glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
		glfwWindowHint(GLFW_RESIZABLE, GLFW_FALSE);
		#if _DEBUG
		glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GL_TRUE);
		#endif

		auto window = glfwCreateWindow(width, height, title.c_str(), NULL, NULL);
		if (!window)
		{
			std::cerr << "Coudld not create window context" << std::endl;
			glfwTerminate();
			return nullptr;
		}

		return std::unique_ptr<Window>(new Window(window));
	}


	GLFWwindow* handle() const {
		return window;
	}

	static Window& getFromHandle(GLFWwindow* handle) {
		return *static_cast<Window*>(glfwGetWindowUserPointer(handle));
	}

	virtual void bind() const override {
		glfwMakeContextCurrent(window);
	}

	bool keepRunning() const {
		return !glfwWindowShouldClose(window);
	}

	bool keyState(int key) const {
		return glfwGetKey(window, key) == GLFW_PRESS;
	}

	bool mouseButtonState(int button) const {
		return glfwGetMouseButton(window, button) == GLFW_PRESS;
	}

	void swapBuffers() {
		glfwSwapBuffers(window);
	}

	void setKeyCallback(std::function<void(Window&, int, int, int, int)> func) {
		keyCallbackFunction = std::move(func);
		glfwSetKeyCallback(window, [](GLFWwindow* handle, int key, int scancode, int action, int mods) {
			auto& window = getFromHandle(handle);
			window.keyCallbackFunction(window, key, scancode, action, mods);
			});
	}

	void setMouseCallback(std::function<void(Window&, double, double)> func) {
		cursorCallbackFunction = std::move(func);
		glfwSetCursorPosCallback(window, [](GLFWwindow* handle, double xpos, double ypos) {
			auto& window = getFromHandle(handle);
			window.cursorCallbackFunction(window, xpos, ypos);
			});
	}

	void setScrollCallback(std::function<void(Window&, double, double)> func) {
		scrollCallbackFunction = std::move(func);
		glfwSetScrollCallback(window, [](GLFWwindow* handle, double xoff, double yoff) {
			auto& window = getFromHandle(handle);
			window.scrollCallbackFunction(window, xoff, yoff);
			});
	}

	void setCloseFlag() {
		glfwSetWindowShouldClose(window, GLFW_TRUE);
	}

	virtual ~Window() {
		glfwDestroyWindow(window);
	}

protected:
	Window(GLFWwindow* w)
		: window(w) {
		glfwSetWindowUserPointer(window, this);
	}

private:
	GLFWwindow* window;
	std::function<void(Window&, int, int, int, int)> keyCallbackFunction;
	std::function<void(Window&, double, double)> cursorCallbackFunction;
	std::function<void(Window&, double, double)> scrollCallbackFunction;
};

#endif
